							;@ ARM flags
	.equ PSR_S, 0x00000008		;@ Sign (negative)
	.equ PSR_Z, 0x00000004		;@ Zero
	.equ PSR_C, 0x00000002		;@ Carry
	.equ PSR_V, 0x00000001		;@ Overflow/Parity
	.equ PSR_P, 0x00000001		;@ Overflow/Parity

	.equ PSR_n, 0x00000080		;@ Was the last opcode add or sub?
	.equ PSR_X, 0x00000040		;@ z80_X (unused)
	.equ PSR_Y, 0x00000020		;@ z80_Y (unused)
	.equ PSR_H, 0x00000010		;@ Half carry


							;@ Z80 flags
	.equ SF, 0x80				;@ Sign (negative)
	.equ ZF, 0x40				;@ Zero
	.equ YF, 0x20				;@ z80_Y (unused)
	.equ HF, 0x10				;@ Half carry
	.equ XF, 0x08				;@ z80_X (unused)
	.equ PF, 0x04				;@ Overflow/Parity
	.equ VF, 0x04				;@ Overflow/Parity
	.equ NF, 0x02				;@ Was the last opcode add or sub?
	.equ CF, 0x01				;@ Carry

	z80f		.req r3			;@
	z80a		.req r4			;@ bits 0-23=0
	z80bc		.req r5			;@ bits 0-15=0
	z80de		.req r6			;@ bits 0-15=0
	z80hl		.req r7			;@ bits 0-15=0
	cycles		.req r8
	z80pc		.req r9
	z80optbl	.req r10
	globalptr	.req r10
	z80sp		.req r11		;@ bits 0-15=0
	z80xy		.req lr			;@ pointer to IX or IY reg
	addy		.req r12		;@ keep this at r12 (scratch for APCS)

;@----------------------------------------------------------------------------
	.equ NOCPUHACK, 2			;@ don't use JMP hack
;@----------------------------------------------------------------------------
	.equ CYC_SHIFT, 8
	.equ CYCLE, 1<<CYC_SHIFT	;@ one cycle
	.equ CYC_MASK, CYCLE-1		;@ Mask
;@----------------------------------------------------------------------------

	.equ z80_opz,			0
	.equ z80_pzst,			z80_opz + 256*4
	.equ z80_regs,			z80_pzst + 256
	.equ z80_regs2,			z80_regs + 8*4
	.equ z80_ix,			z80_regs2 + 8*4
	.equ z80_iy,			z80_ix + 4
	.equ z80_i,				z80_iy + 4
	.equ z80_r,				z80_i + 1
	.equ z80_im,			z80_r + 1
	.equ z80_iff2,			z80_im + 1
	.equ z80_iff1,			z80_iff2 + 1
	.equ z80_irq_pin,		z80_iff1 + 1
	.equ z80_temp1,			z80_irq_pin + 1
	.equ z80_temp2,			z80_temp1 + 1
	.equ z80_nmi_pin,		z80_temp2 + 1
	.equ z80_old_nmi,		z80_nmi_pin + 1
	.equ z80_out0,			z80_old_nmi + 1
	.equ z80_temp3,			z80_out0 + 1
	.equ z80_lastbank,		z80_temp3 + 1
	.equ z80_nexttimeout,	z80_lastbank + 4
	.equ z80_nexttimeout_,	z80_nexttimeout + 4
	.equ z80_oldcycles,		z80_nexttimeout_ + 4


	.macro get_pzst_tbl x
	add \x,z80optbl,#z80_pzst
	.endm

	.macro load_lastbank x
	ldr \x,[z80optbl,#z80_lastbank]
	.endm

	.macro store_lastbank x
	str \x,[z80optbl,#z80_lastbank]
	.endm

	.macro encodePC				;@ translate z80pc from Z80 PC to rom offset
	bl translatePCToOffset		;@ in=z80pc, out=r0
	store_lastbank r0
	add z80pc,z80pc,r0
	.endm

	.macro encodePC_IRQ_			;@ translate z80pc from zeropage Z80 PC to rom offset
	get_memmap_tbl r1
	ldr r1,[r1]
	store_lastbank r1
	add z80pc,addy,r1
	.endm

	.macro reencodePC			;@ translate z80pc from Z80 PC to rom offset
	load_lastbank r0
	sub z80pc,z80pc,r0
	encodePC
	.endm

	.macro encodeFLG			;@ pack Z80 flags into r0
	and r0,z80f,#PSR_H|PSR_Y
	and r1,z80f,#PSR_S|PSR_Z
	orr r0,r0,r1,lsl#4
	movs r1,z80f,lsl#31
	orrmi r0,r0,#VF
	and r1,z80f,#PSR_n
	adc r0,r0,r1,lsr#6			;@ NF & CF
	tst z80f,#PSR_X
	orrne r0,r0,#XF
	.endm

	.macro decodeFLG			;@ unpack Z80 flags from r0
	and z80f,r0,#HF|YF
	tst r0,#XF
	orrne z80f,z80f,#PSR_X
	tst r0,#CF
	orrne z80f,z80f,#PSR_C
	and r1,r0,#SF|ZF
	movs r0,r0,lsl#30
	adc z80f,z80f,r1,lsr#4		;@ also sets V/P Flag.
	orrmi z80f,z80f,#PSR_n
	.endm


/*	.macro fetch count
	subs cycles,cycles,#\count*CYCLE
	b fetchdebug
	.endm
*/
	.macro fetch count
	subs cycles,cycles,#\count*CYCLE
	ldrplb r0,[z80pc],#1
	ldrpl pc,[z80optbl,r0,lsl#2]
	ldr pc,[z80optbl,#z80_nexttimeout]
	.endm

	.macro fetch_force
	ldrb r0,[z80pc],#1
	ldr pc,[z80optbl,r0,lsl#2]
	.endm

	.macro eatcycles count
	sub cycles,cycles,#\count*CYCLE
	.endm

	.macro readmem8
	bl mem_R8
	.endm

	.macro readmem8BC
	mov addy,z80bc,lsr#16
	readmem8
	.endm

	.macro readmem8DE
	mov addy,z80de,lsr#16
	readmem8
	.endm

	.macro readmem8HL
	mov addy,z80hl,lsr#16
	readmem8
	.endm

	.macro readmem16 reg
	readmem8
	mov \reg,r0,lsl#16
	add addy,addy,#1
	readmem8
	orr \reg,\reg,r0,lsl#24
	.endm

	.macro writemem8
	bl mem_W8
	.endm

	.macro writemem8BC
	mov addy,z80bc,lsr#16
	writemem8
	.endm

	.macro writemem8DE
	mov addy,z80de,lsr#16
	writemem8
	.endm

	.macro writemem8HL
	mov addy,z80hl,lsr#16
	writemem8
	.endm

	.macro writemem16 reg
	mov r0,\reg,lsr#16
	writemem8
	add addy,addy,#1
	mov r0,\reg,lsr#24
	writemem8
	.endm

	.macro copymem8HL_DE
	readmem8HL
	writemem8DE
	.endm

	.macro CalcIXd
	ldrsb r1,[z80pc],#1
	ldr addy,[z80xy]
	add addy,addy,r1,lsl#16
	mov addy,addy,lsr#16
	.endm
;@----------------------------------------------------------------------------

	.macro push16				;@ push r0
	sub z80sp,z80sp,#0x00010000
	mov addy,z80sp,lsr#16
	stmfd sp!,{r0}
	mov r0,r0,lsr#8
	writemem8
	sub z80sp,z80sp,#0x00010000
	mov addy,z80sp,lsr#16
	ldmfd sp!,{r0}
	writemem8
	.endm						;@ r1,r2=?

	.macro push16Reg x			;@ push BC,DE,HL
	sub z80sp,z80sp,#0x00010000
	mov addy,z80sp,lsr#16
	mov r0,\x,lsr#24
	writemem8
	sub z80sp,z80sp,#0x00010000
	mov addy,z80sp,lsr#16
	mov r0,\x,lsr#16
	writemem8
	.endm						;@ r1,r2=?

	.macro pop16Reg x			;@ pop BC,DE,HL
	mov addy,z80sp,lsr#16
	readmem8
	mov \x,r0,lsl#16
	add z80sp,z80sp,#0x00010000
	mov addy,z80sp,lsr#16
	readmem8
	orr \x,\x,r0,lsl#24
	add z80sp,z80sp,#0x00010000
	.endm						;@ r0,r1=?

	.macro pop16AF				;@ pop AF
	mov addy,z80sp,lsr#16
	readmem8
	mov z80f,r0
	add z80sp,z80sp,#0x00010000
	mov addy,z80sp,lsr#16
	readmem8
	mov z80a,r0,lsl#24
	add z80sp,z80sp,#0x00010000
	mov r0,z80f
	.endm						;@ r0=flags,r1=?

	.macro pop16PC				;@ pop PC
	mov addy,z80sp,lsr#16
	readmem8
	mov z80pc,r0
	add z80sp,z80sp,#0x00010000
	mov addy,z80sp,lsr#16
	readmem8
	orr z80pc,z80pc,r0,lsl#8
	add z80sp,z80sp,#0x00010000
	.endm						;@ r0,r1=?

	.macro pop16IX				;@ pop IX
	stmfd sp!,{z80xy}
	mov addy,z80sp,lsr#16
	readmem8
	stmfd sp!,{r0}
	add z80sp,z80sp,#0x00010000
	mov addy,z80sp,lsr#16
	readmem8
	ldmfd sp!,{r1}
	orr r0,r1,r0,lsl#8
	add z80sp,z80sp,#0x00010000
	ldmfd sp!,{z80xy}
	strh r0,[z80xy,#2]
	.endm						;@ r0,r1=?
;@----------------------------------------------------------------------------

	.macro opADC
	movs z80f,z80f,lsr#2		;@ get C
	subcs r0,r0,#0x100
	eor z80f,r0,z80a,lsr#24		;@ prepare for check of half carry
	adcs z80a,z80a,r0,ror#8
	mrs r0,cpsr					;@ S,Z,V&C
	eor z80f,z80f,z80a,lsr#24
	and z80f,z80f,#PSR_H		;@ H, correct
	orr z80f,z80f,r0,lsr#28
	.endm

	.macro opADCA
	movs z80f,z80f,lsr#2		;@ get C
	orrcs z80a,z80a,#0x00800000
	adds z80a,z80a,z80a
	mrs z80f,cpsr				;@ S,Z,V&C
	mov z80f,z80f,lsr#28
	tst z80a,#0x10000000		;@ H, correct
	orrne z80f,z80f,#PSR_H
	fetch 4
	.endm

	.macro opADCH x
	mov r0,\x,lsr#24
	opADC
	fetch 4
	.endm

	.macro opADCL x
	movs z80f,z80f,lsr#2		;@ get C
	adc r0,\x,\x,lsr#15
	orrcs z80a,z80a,#0x00800000
	mov r1,z80a,lsl#4			;@ Prepare for check of half carry
	adds z80a,z80a,r0,lsl#23
	mrs z80f,cpsr				;@ S,Z,V&C
	mov z80f,z80f,lsr#28
	cmn r1,r0,lsl#27
	orrcs z80f,z80f,#PSR_H		;@ H, correct
	fetch 4
	.endm

	.macro opADCb
	opADC
	.endm
;@---------------------------------------

	.macro opADD x y
	mov r1,z80a,lsl#4			;@ Prepare for check of half carry
	adds z80a,z80a,\x,lsl#\y
	mrs z80f,cpsr				;@ S,Z,V&C
	mov z80f,z80f,lsr#28
	cmn r1,\x,lsl#\y+4
	orrcs z80f,z80f,#PSR_H
	.endm

	.macro opADDA
	adds z80a,z80a,z80a
	mrs z80f,cpsr				;@ S,Z,V&C
	mov z80f,z80f,lsr#28
	tst z80a,#0x10000000		;@ H, correct
	orrne z80f,z80f,#PSR_H
	fetch 4
	.endm

	.macro opADDH x
	and r0,\x,#0xFF000000
	opADD r0,0
	fetch 4
	.endm

	.macro opADDL x
	opADD \x,8
	fetch 4
	.endm

	.macro opADDb 
	opADD r0,24
	.endm
;@---------------------------------------

	.macro opADC16 x
	movs z80f,z80f,lsr#2		;@ get C
	adc r0,z80a,\x,lsr#15
	orrcs z80hl,z80hl,#0x00008000
	mov r1,z80hl,lsl#4
	adds z80hl,z80hl,r0,lsl#15
	mrs z80f,cpsr				;@ S,Z,V&C
	mov z80f,z80f,lsr#28
	cmn r1,r0,lsl#19
	orrcs z80f,z80f,#PSR_H
	fetch 15
	.endm

	.macro opADC16HL
	movs z80f,z80f,lsr#2		;@ get C
	orrcs z80hl,z80hl,#0x00008000
	adds z80hl,z80hl,z80hl
	mrs z80f,cpsr				;@ S,Z,V&C
	mov z80f,z80f,lsr#28
	tst z80hl,#0x10000000		;@ H, correct.
	orrne z80f,z80f,#PSR_H
	fetch 15
	.endm

	.macro opADD16 y x
	mov r1,\y,lsl#4				;@ Prepare for check of half carry
	adds \y,\y,\x
	bic z80f,z80f,#PSR_C+PSR_H+PSR_n
	orrcs z80f,z80f,#PSR_C
	cmn r1,\x,lsl#4
	orrcs z80f,z80f,#PSR_H
	.endm

	.macro opADD16_2 x
	adds \x,\x,\x
	bic z80f,z80f,#PSR_C+PSR_H+PSR_n
	orrcs z80f,z80f,#PSR_C
	tst \x,#0x10000000			;@ H, correct.
	orrne z80f,z80f,#PSR_H
	.endm
;@---------------------------------------

	.macro opAND x y
	and z80a,z80a,\x,lsl#\y
	get_pzst_tbl r0
	ldrb z80f,[r0,z80a,lsr#24]	;@ get PZS
	orr z80f,z80f,#PSR_H		;@ set PSR_H
	.endm

	.macro opANDA
	get_pzst_tbl r0
	ldrb z80f,[r0,z80a,lsr#24]	;@ get PZS
	orr z80f,z80f,#PSR_H		;@ set PSR_H
	fetch 4
	.endm

	.macro opANDH x
	opAND \x,0
	fetch 4
	.endm

	.macro opANDL x
	opAND \x,8
	fetch 4
	.endm

	.macro opANDb
	opAND r0,24
	.endm
;@---------------------------------------

	.macro opBIT x
	and z80f,z80f,#PSR_C			;@ keep C
	orr z80f,z80f,#PSR_H			;@ set H
	mov r0,r0,lsr#3
	tst \x,r1,lsl r0				;@ r0 0x08-0x0F
	orreq z80f,z80f,#PSR_Z|PSR_P	;@ Z & P
	fetch 8
	.endm

	.macro opBITH x
	mov r1,#0x00010000
	opBIT \x
	.endm

	.macro opBITL x
	mov r1,#0x00000100
	opBIT \x
	.endm

	.macro opBIT7H x
	and z80f,z80f,#PSR_C				;@ keep C
	tst \x,#0x80000000					;@ bit 7
	orreq z80f,z80f,#PSR_H+PSR_Z+PSR_P	;@ H,Z & P
	orrne z80f,z80f,#PSR_H+PSR_S		;@ H & sign on "BIT 7,x"
	fetch 8
	.endm

	.macro opBIT7L x
	and z80f,z80f,#PSR_C				;@ keep C
	tst \x,#0x00800000					;@ bit 7
	orreq z80f,z80f,#PSR_H+PSR_Z+PSR_P	;@ H,Z & P
	orrne z80f,z80f,#PSR_H+PSR_S		;@ H & sign on "BIT 7,x"
	fetch 8
	.endm

	.macro opBITmem x
	readmem8
	and z80f,z80f,#PSR_C			;@ keep C
	orr z80f,z80f,#PSR_H			;@ set H
	tst r0,#1<<\x					;@ bit x
	orreq z80f,z80f,#PSR_Z|PSR_P	;@ Z & P
	fetch 12
	.endm

	.macro opBIT7mem
	readmem8
	and z80f,z80f,#PSR_C				;@ keep C
	tst r0,#0x80						;@ bit x
	orreq z80f,z80f,#PSR_H+PSR_Z+PSR_P	;@ H,Z & P
	orrne z80f,z80f,#PSR_H+PSR_S		;@ H & sign on "BIT 7,x"
	fetch 12
	.endm
;@---------------------------------------

	.macro opCP x y
	mov r1,z80a,lsl#4			;@ prepare for check of half carry
	cmp z80a,\x,lsl#\y
	mrs z80f,cpsr				;@ S,Z,V&C
	mov z80f,z80f,lsr#28
	eor z80f,z80f,#PSR_C|PSR_n	;@ invert C and set n
	cmp r1,\x,lsl#\y+4
	orrcc z80f,z80f,#PSR_H
	.endm

	.macro opCPA
	mov z80f,#PSR_Z|PSR_n		;@ set Z & n
	fetch 4
	.endm

	.macro opCPH x
	and r0,\x,#0xFF000000
	opCP r0,0
	fetch 4
	.endm

	.macro opCPL x
	opCP \x,8
	fetch 4
	.endm

	.macro opCPb
	opCP r0,24
	.endm
;@---------------------------------------

	.macro opDEC8 x				;@ for A and memory
	orr z80f,z80f,#PSR_n+PSR_H+PSR_S+PSR_V+PSR_Z	;@ save carry & set n
	tst \x,#0x0f000000
	bicne z80f,z80f,#PSR_H
	subs \x,\x,#0x01000000
	bicpl z80f,z80f,#PSR_S
	bicvc z80f,z80f,#PSR_V
	bicne z80f,z80f,#PSR_Z
	.endm

	.macro opDEC8H x			;@ for B, D & H
	orr z80f,z80f,#PSR_n+PSR_H+PSR_S+PSR_V+PSR_Z	;@ save carry & set n
	tst \x,#0x0f000000
	bicne z80f,z80f,#PSR_H
	subs \x,\x,#0x01000000
	bicpl z80f,z80f,#PSR_S
	bicvc z80f,z80f,#PSR_V
	tst \x,#0xff000000			;@ Z
	bicne z80f,z80f,#PSR_Z
	.endm

	.macro opDEC8L x			;@ for C, E & L
	mov \x,\x,ror#24
	opDEC8H \x
	mov \x,\x,ror#8
	.endm

	.macro opDEC8b				;@ for memory
	mov r0,r0,lsl#24
	opDEC8 r0
	mov r0,r0,lsr#24
	.endm

	.macro opDEC16 x
	sub \x,\x,#0x00010000
	.endm
;@---------------------------------------

	.macro opINC8 x
	and z80f,z80f,#PSR_C		;@ save carry, clear n
	adds \x,\x,#0x01000000
	orrmi z80f,z80f,#PSR_S
	orrvs z80f,z80f,#PSR_V
	orrcs z80f,z80f,#PSR_Z		;@ when going from 0xFF to 0x00 carry is set.
	tst \x,#0x0f000000			;@ h
	orreq z80f,z80f,#PSR_H
	.endm

	.macro opINC8H x			;@ for B, D & H
	opINC8 \x
	.endm

	.macro opINC8L x			;@ for C, E & L
	mov \x,\x,ror#24
	opINC8 \x
	mov \x,\x,ror#8
	.endm

	.macro opINC8b				;@ for memory
	mov r0,r0,lsl#24
	opINC8 r0
	mov r0,r0,lsr#24
	.endm

	.macro opINC16 x
	add \x,\x,#0x00010000
	.endm
;@---------------------------------------

	.macro opINrC
	bl Z80_IN_C					;@ uses z80bc
	get_pzst_tbl r1
	ldrb r1,[r1,r0]				;@ get PZS
	and z80f,z80f,#PSR_C		;@ keep C
	orr z80f,z80f,r1
	.endm
;@---------------------------------------

	.macro opLDIM16
	ldrb r0,[z80pc],#1
	ldrb r1,[z80pc],#1
	orr r0,r0,r1,lsl#8
	.endm

	.macro opLDIM8H x
	ldrb r0,[z80pc],#1
	and \x,\x,#0x00ff0000
	orr \x,\x,r0,lsl#24
	.endm

	.macro opLDIM8L x
	ldrb r0,[z80pc],#1
	and \x,\x,#0xff000000
	orr \x,\x,r0,lsl#16
	.endm
;@---------------------------------------

	.macro opOR x y
	orr z80a,z80a,\x,lsl#\y
	get_pzst_tbl r1
	ldrb z80f,[r1,z80a,lsr#24]	;@ get PZS
	.endm

	.macro opORA
	get_pzst_tbl r1
	ldrb z80f,[r1,z80a,lsr#24]	;@ get PZS
	fetch 4
	.endm

	.macro opORH x
	and r0,\x,#0xFF000000
	opOR r0,0
	fetch 4
	.endm

	.macro opORL x
	opOR \x,8
	fetch 4
	.endm

	.macro opORb
	opOR r0,24
	.endm
;@---------------------------------------

	.macro opOUTCr
	bl Z80_OUT_C
	fetch 12
	.endm

	.macro opOUTCrH x
	mov r0,\x,lsr#24
	opOUTCr
	.endm

	.macro opOUTCrL x
	mov r0,\x,lsr#16
;@	and r0,r0,#0xFF
	opOUTCr
	.endm
;@---------------------------------------

	.macro opRES x
	mov r0,r0,lsr#3
	bic \x,\x,r1,lsl r0			;@ r0 0x10-0x17
	.endm

	.macro opRESH x
	mov r1,#0x00000100
	opRES \x
	fetch 8
	.endm

	.macro opRESL x
	mov r1,#0x00000001
	opRES \x
	fetch 8
	.endm

	.macro opRESmem x
	readmem8
	bic r0,r0,#1<<\x			;@ bit ?
	writemem8
	fetch 15
	.endm
;@---------------------------------------

	.macro opRL x y z
	movs \x,\y,lsl#\z
	tst z80f,#PSR_C				;@ doesn't affect ARM carry, as long as the imidiate value is < 0x100. Watch out!
	orrne \x,\x,#0x01000000
;@	and r2,z80f,#PSR_C
;@	orr \x,\x,r2,lsl#23
	get_pzst_tbl r1
	ldrb z80f,[r1,\x,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	.endm

	.macro opRLA
	opRL z80a,z80a,1
	fetch 8
	.endm

	.macro opRLH x
	and r0,\x,#0xFF000000		;@ mask high to r0
	adds \x,\x,r0
	tst z80f,#PSR_C				;@ doesn't affect ARM carry, as long as the imidiate value is < 0x100. Watch out!
	orrne \x,\x,#0x01000000
	get_pzst_tbl r1
	ldrb z80f,[r1,\x,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	fetch 8
	.endm

	.macro opRLL x
	opRL r0,\x,9
	and \x,\x,#0xFF000000		;@ mask out high
	orr \x,\x,r0,lsr#8
	fetch 8
	.endm

	.macro opRLb
	opRL r0,r0,25
	mov r0,r0,lsr#24
	.endm
;@---------------------------------------

	.macro opRLC x y z
	movs \x,\y,lsl#\z
	orrcs \x,\x,#0x01000000
	get_pzst_tbl r1
	ldrb z80f,[r1,\x,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	.endm

	.macro opRLCA
	opRLC z80a,z80a,1
	fetch 8
	.endm

	.macro opRLCH x
	and r0,\x,#0xFF000000		;@ mask high to r0
	adds \x,\x,r0
	orrcs \x,\x,#0x01000000
	get_pzst_tbl r1
	ldrb z80f,[r1,\x,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	fetch 8
	.endm

	.macro opRLCL x
	opRLC r0,\x,9
	and \x,\x,#0xFF000000		;@ mask out high
	orr \x,\x,r0,lsr#8
	fetch 8
	.endm

	.macro opRLCb
	opRLC r0,r0,25
	mov r0,r0,lsr#24
	.endm
;@---------------------------------------

	.macro opRR x y z
	movs \x,\y,lsr#\z
	tst z80f,#PSR_C				;@ doesn't affect ARM carry, as long as the imidiate value is < 0x100. Watch out!
	orrne \x,\x,#0x00000080
;@	and r1,z80f,#PSR_C
;@	orr \x,\x,r1,lsl#6
	get_pzst_tbl r1
	ldrb z80f,[r1,\x]			;@ get PZS
	orrcs z80f,z80f,#PSR_C
	.endm

	.macro opRRA
	orr z80a,z80a,z80f,lsr#1	;@ get C
	movs z80a,z80a,ror#25
	mov z80a,z80a,lsl#24
	get_pzst_tbl r1
	ldrb z80f,[r1,z80a,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	fetch 8
	.endm

	.macro opRRH x
	orr r0,\x,z80f,lsr#1		;@ get C
	movs r0,r0,ror#25
	and \x,\x,#0x00FF0000		;@ mask out low
	orr \x,\x,r0,lsl#24
	get_pzst_tbl r1
	ldrb z80f,[r1,\x,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	fetch 8
	.endm

	.macro opRRL x
	and r0,\x,#0x00FF0000		;@ mask out low to r0
	opRR r0,r0,17
	and \x,\x,#0xFF000000		;@ mask out high
	orr \x,\x,r0,lsl#16
	fetch 8
	.endm

	.macro opRRb
	opRR r0,r0,1
	.endm
;@---------------------------------------

	.macro opRRC x y z
	movs \x,\y,lsr#\z
	orrcs \x,\x,#0x00000080
	get_pzst_tbl r1
	ldrb z80f,[r1,\x]			;@ get PZS
	orrcs z80f,z80f,#PSR_C
	.endm

	.macro opRRCA
	opRRC z80a,z80a,25
	mov z80a,z80a,lsl#24
	fetch 8
	.endm

	.macro opRRCH x
	opRRC r0,\x,25
	and \x,\x,#0x00FF0000		;@ mask out low
	orr \x,\x,r0,lsl#24
	fetch 8
	.endm

	.macro opRRCL x
	and r0,\x,#0x00FF0000		;@ mask low to r0
	opRRC r0,r0,17
	and \x,\x,#0xFF000000		;@ mask out high
	orr \x,\x,r0,lsl#16
	fetch 8
	.endm

	.macro opRRCb
	opRRC r0,r0,1
	.endm
;@---------------------------------------

	.macro opSBC
	and z80f,z80f,#PSR_C
	subs r0,r0,z80f,lsl#7		;@ fix up r0 and set correct C.
	eor z80f,r0,z80a,lsr#24		;@ prepare for check of H
	sbcs z80a,z80a,r0,ror#8
	mrs r0,cpsr
	eor z80f,z80f,z80a,lsr#24
	and z80f,z80f,#PSR_H		;@ H, correct
	orr z80f,z80f,r0,lsr#28		;@ S,Z,V&C
	eor z80f,z80f,#PSR_C|PSR_n	;@ invert C and set n.
	.endm

	.macro opSBCA
	ands z80a,z80f,#PSR_C		;@ clear z80a, check C
	movne z80a,#0xFF000000
	moveq z80f,#PSR_n+PSR_Z
	movne z80f,#PSR_n+PSR_S+PSR_C+PSR_H
	fetch 4
	.endm

	.macro opSBCH x
	mov r0,\x,lsr#24
	opSBC
	fetch 4
	.endm

	.macro opSBCL x
	mov r0,\x,lsl#8
	eor z80f,z80f,#PSR_C		;@ invert C
	movs z80f,z80f,lsr#2		;@ get C
	sbccc r0,r0,#0xFF000000
	mov r1,z80a,lsl#4			;@ prepare for check of H
	sbcs z80a,z80a,r0
	mrs z80f,cpsr				;@ S,Z,V&C
	mov z80f,z80f,lsr#28
	eor z80f,z80f,#PSR_C|PSR_n	;@ invert C and set n.
	cmp r1,r0,lsl#4
	orrcc z80f,z80f,#PSR_H		;@ H, correct
	fetch 4
	.endm

	.macro opSBCb
	opSBC
	.endm
;@---------------------------------------

	.macro opSBC16 x
	and z80f,z80f,#PSR_C
	subs r1,z80f,z80f,lsl#1		;@ fix up r1 and set correct C.
	orr r0,\x,r1,lsr#16
	mov r1,z80hl,lsl#4			;@ prepare for check of H
	sbcs z80hl,z80hl,r0
	mrs z80f,cpsr				;@ S,Z,V&C
	mov z80f,z80f,lsr#28
	eor z80f,z80f,#PSR_C|PSR_n	;@ invert C and set n.
	cmp r1,r0,lsl#4
	orrcc z80f,z80f,#PSR_H
	fetch 15
	.endm

	.macro opSBC16HL
	ands z80hl,z80f,z80f,lsl#31	;@ clear z80hl, check C
	subcs z80hl,z80hl,#0x00010000
	movcc z80f,#PSR_n+PSR_Z
	movcs z80f,#PSR_n+PSR_S+PSR_C+PSR_H
	fetch 15
	.endm
;@---------------------------------------

	.macro opSET x
	mov r0,r0,lsr#3
	and r0,r0,#7
	orr \x,\x,r1,lsl r0			;@ r0 0-7
	.endm

	.macro opSETH x
	mov r1,#0x01000000
	opSET \x
	fetch 8
	.endm

	.macro opSETL x
	mov r1,#0x00010000
	opSET \x
	fetch 8
	.endm

	.macro opSETmem x
	readmem8
	orr r0,r0,#1<<\x			;@ bit ?
	writemem8
	fetch 15
	.endm
;@---------------------------------------

	.macro opSLA x y z
	movs \x,\y,lsl#\z
	get_pzst_tbl r1
	ldrb z80f,[r1,\x,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	.endm

	.macro opSLAA
	opSLA z80a,z80a,1
	fetch 8
	.endm

	.macro opSLAH x
	and r0,\x,#0xFF000000		;@ mask high to r0
	adds \x,\x,r0
	get_pzst_tbl r1
	ldrb z80f,[r1,\x,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	fetch 8
	.endm

	.macro opSLAL x
	opSLA r0,\x,9
	and \x,\x,#0xFF000000		;@ mask out high
	orr \x,\x,r0,lsr#8
	fetch 8
	.endm

	.macro opSLAb
	opSLA r0,r0,25
	mov r0,r0,lsr#24
	.endm
;@---------------------------------------

	.macro opSLL x y z
	movs \x,\y,lsl#\z
	orr \x,\x,#0x01000000
	get_pzst_tbl r1
	ldrb z80f,[r1,\x,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	.endm

	.macro opSLLA
	opSLL z80a,z80a,1
	fetch 8
	.endm

	.macro opSLLH x
	and r0,\x,#0xFF000000		;@ mask high to r0
	adds \x,\x,r0
	orr \x,\x,#0x01000000
	get_pzst_tbl r1
	ldrb z80f,[r1,\x,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	fetch 8
	.endm

	.macro opSLLL x
	opSLL r0,\x,9
	and \x,\x,#0xFF000000		;@ mask out high
	orr \x,\x,r0,lsr#8
	fetch 8
	.endm

	.macro opSLLb
	opSLL r0,r0,25
	mov r0,r0,lsr#24
	.endm
;@---------------------------------------

	.macro opSRA x y
	movs \x,\y,asr#25
	and \x,\x,#0xFF
	get_pzst_tbl r1
	ldrb z80f,[r1,\x]			;@ get PZS
	orrcs z80f,z80f,#PSR_C
	.endm

	.macro opSRAA
	movs z80a,z80a,asr#25
	mov z80a,z80a,lsl#24
	get_pzst_tbl r1
	ldrb z80f,[r1,z80a,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	fetch 8
	.endm

	.macro opSRAH x
	movs r0,\x,asr#25
	and \x,\x,#0x00FF0000		;@ mask out low
	orr \x,\x,r0,lsl#24
	get_pzst_tbl r1
	ldrb z80f,[r1,\x,lsr#24]	;@ get PZS
	orrcs z80f,z80f,#PSR_C
	fetch 8
	.endm

	.macro opSRAL x
	mov r0,\x,lsl#8
	opSRA r0,r0
	and \x,\x,#0xFF000000		;@ mask out high
	orr \x,\x,r0,lsl#16
	fetch 8
	.endm

	.macro opSRAb
	and r1,r0,#0x80
	orrs r0,r1,r0,lsr#1
	get_pzst_tbl r1
	ldrb z80f,[r1,r0]			;@ get PZS
	orrcs z80f,z80f,#PSR_C
	.endm
;@---------------------------------------

	.macro opSRL x y z
	movs \x,\y,lsr#\z
	get_pzst_tbl r1
	ldrb z80f,[r1,\x]			;@ get PZS
	orrcs z80f,z80f,#PSR_C
	.endm

	.macro opSRLA
	opSRL z80a,z80a,25
	mov z80a,z80a,lsl#24
	fetch 8
	.endm

	.macro opSRLH x
	opSRL r0,\x,25
	and \x,\x,#0x00FF0000		;@ mask out low
	orr \x,\x,r0,lsl#24
	fetch 8
	.endm

	.macro opSRLL x
	mov r0,\x,lsl#8
	opSRL r0,r0,25
	and \x,\x,#0xFF000000		;@ mask out high
	orr \x,\x,r0,lsl#16
	fetch 8
	.endm

	.macro opSRLb
	opSRL r0,r0,1
	.endm
;@---------------------------------------

	.macro opSUB x y
	mov r1,z80a,lsl#4 			;@ Prepare for check of half carry
	subs z80a,z80a,\x,lsl#\y
	mrs z80f,cpsr				;@ S,Z,V&C
	mov z80f,z80f,lsr#28
	eor z80f,z80f,#PSR_C|PSR_n	;@ invert C and set n
	cmp r1,\x,lsl#\y+4
	orrcc z80f,z80f,#PSR_H
	.endm

	.macro opSUBA
	mov z80a,#0
	mov z80f,#PSR_Z|PSR_n		;@ set Z & n
	fetch 4
	.endm

	.macro opSUBH x
	and r0,\x,#0xFF000000
	opSUB r0,0
	fetch 4
	.endm

	.macro opSUBL x
	opSUB \x,8
	fetch 4
	.endm

	.macro opSUBb
	opSUB r0,24
	.endm
;@---------------------------------------

	.macro opXOR x y
	eor z80a,z80a,\x,lsl#\y
	get_pzst_tbl r0
	ldrb z80f,[r0,z80a,lsr#24]	;@ get PZS
	.endm

	.macro opXORA
	mov z80a,#0					;@ clear A.
	mov z80f,#PSR_Z|PSR_P		;@ Z & P
	fetch 4
	.endm

	.macro opXORH x
	and r0,\x,#0xFF000000
	opXOR r0,0
	fetch 4
	.endm

	.macro opXORL x
	opXOR \x,8
	fetch 4
	.endm

	.macro opXORb
	opXOR r0,24
	.endm
;@---------------------------------------

