#include <nds.h>
#include <fat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
//#include <dirent.h>
#include <sys/dir.h>

#include "gui.h"
#include "filehandling.h"
#include "romloading.h"
#include "unzip/unzipnds.h"
#include "equates.h"

#define FOLDERNAME "s8ds"
#define FILEEXTENSIONS ".sms.gg.sg.sc"
#define SETTINGNAME "settings.cfg"
#define STATESIZE (0x8000+0x2000+0x4080+8+104+80+60+24)

//asm variables
extern u8 g_flicker;			// from gfx.s
extern u8 g_gammavalue;			// from gfx.s
extern u32 g_ROM_Size;			// cart.s
extern u32 g_emuflags;			// cart.s
extern u8 g_config_set;			// cart.s
extern u8 g_scaling_set;		// cart.s
extern u32 joycfg;				// io.s

//asm functions
extern void loadcart(int,int);		// cart.s
extern int packState(u32* statePtr);	// cart.s
extern void unpackState(u32* statePtr);	// cart.s
extern void CPU_reset(void);		// cpu.s

extern u8 ROM_Space[];				// cart.s
extern u8 BIOS_US_Space[];			// cart.s
extern u8 BIOS_JP_Space[];			// cart.s
extern u8 BIOS_GG_Space[];			// cart.s
extern void * g_BIOSBASE_US;		// cart.s
extern void * g_BIOSBASE_JP;		// cart.s
extern void * g_BIOSBASE_GG;		// cart.s
extern int sleepTime;				// main.c
extern int selectedGame;			// main.c
extern int powerButton;				// main.c
extern int pauseEmulation;			// main.c
extern int emuSettings;				// gui.c

extern void loadcart(int,int);		// cart.s
extern void waitVBlank(void);		// main.c
extern void pausVBlank(int count);	// main.c
extern void setEmuSpeed(int speed);	// main.c


int fatAvailable = 0;
int directoryCached = 0;
char * dTable = 0;
int dItems=0;
configdata cfg;

char tempdir[256];
char currentdir[256];
char currentfilename[64];

char *const spinner[]={"\\","|","/","-"};
//---------------------------------------------------------------------------------
void initFileHandling() {
	strcpy(currentdir,"/");
}


int loadSettings() {
	FILE* myfile;

	cfg.currentPath[0]=0;
	cfg.biosUS[0]=0;
	cfg.biosJP[0]=0;
	cfg.biosGG[0]=0;
	if(findFolder(FOLDERNAME)) {
		pausVBlank(60*3);
		return 1;
	}
	if( (myfile = fopen(SETTINGNAME, "r")) ) {
		fread(&cfg, 1, sizeof(configdata), myfile);
		fclose(myfile);
		if(	!strstr(cfg.magic,"cfg")) {
			drawtext("Error in settings file.", 18, 0);
			pausVBlank(60*3);
			return 1;
		}
	} else {
		drawtext("Couldn't open file:", 18, 0);
		drawtext(SETTINGNAME, 19, 0);
		pausVBlank(60*3);
		return 1;
	}

//	g_dipswitch0 = cfg.dipswitch0;
//	g_dipswitch1 = cfg.dipswitch1;
	g_config_set  = cfg.config;
	g_scaling_set = cfg.scaling&3;
	g_flicker     = cfg.flicker&1;
	g_gammavalue  = cfg.gammavalue;
	emuSettings   = cfg.emuSettings &~ 0xC0;			// Clear speed setting.
	sleepTime     = cfg.sleepTime;
	joycfg        = (joycfg&~0x400)|((cfg.controller&1)<<10);
	strcpy(currentdir, cfg.currentPath);

	drawtext("Settings loaded.", 18, 0);
	return 0;
}

void saveSettings() {
	FILE* myfile;

	strcpy(cfg.magic,"cfg");
//	cfg.dipswitch0  = g_dipswitch0;
//	cfg.dipswitch1  = g_dipswitch1;
	cfg.config      = g_config_set;
	cfg.scaling     = g_scaling_set&3;
	cfg.flicker     = g_flicker&1;
	cfg.gammavalue  = g_gammavalue;
	cfg.emuSettings = emuSettings &~ 0xC0;			// Clear speed setting.
	cfg.sleepTime   = sleepTime;
	cfg.controller  = (joycfg>>10)&1;
	strcpy(cfg.currentPath, currentdir);

	if(findFolder(FOLDERNAME)) {
		pausVBlank(60*3);
		return;
	}
	if( (myfile = fopen(SETTINGNAME, "w")) ) {
		fwrite(&cfg, 1, sizeof(configdata), myfile);
		fclose(myfile);
	} else {
		drawtext("Couldn't open file:", 18, 0);
		drawtext(SETTINGNAME, 19, 0);
		pausVBlank(60*3);
		return;
	}
	drawtext("Settings saved.", 18, 0);
}

void selectSMSGame() {
	const char * smsName = NULL;
	char fileExt[8];

	pauseEmulation = 1;
	cls(0);
	smsName = browseForFileType(FILEEXTENSIONS".zip");
	if(smsName) {
		cls(0);
		drawtext("   Please wait, loading.", 12, 0);
		g_emuflags &=~(MD_MODE|SGAC_MODE);
		g_ROM_Size = loadROM(ROM_Space, smsName, 0x100000);
		if(!g_ROM_Size) {
			g_ROM_Size = loadArcadeROM(ROM_Space, smsName);
		}
		if(g_ROM_Size) {
			g_emuflags &=~(GG_MODE|SG_MODE);
			getFileExtension(fileExt, currentfilename);
			if(strstr(fileExt, ".gg"))
				g_emuflags |=GG_MODE;
			if(strstr(fileExt, ".sg") || strstr(fileExt, ".sc"))
				g_emuflags |=SG_MODE;
			setEmuSpeed(0);
			loadcart(0,g_emuflags);
			if(emuSettings&4)
				loadState();
			powerButton = 1;
		} else {
			pausVBlank(60*2);
		}
	}
	pauseEmulation = emuSettings&1;
}

void selectUSBios() {
	const char * biosName;
	
	directoryCached = 0;
	biosName = browseForFileType(".sms.zip");
	directoryCached = 0;
	if(biosName) {
		strcpy(cfg.biosUS, currentdir);
		strcat(cfg.biosUS,"/");
		strcat(cfg.biosUS,biosName);
		if(loadUSBIOS()) {
			return;
		}
		pausVBlank(60*2);
	}
}

void selectJPBios() {
	const char * biosName;
	
	directoryCached = 0;
	biosName = browseForFileType(".sms.zip");
	directoryCached = 0;
	if(biosName) {
		strcpy(cfg.biosJP, currentdir);
		strcat(cfg.biosJP,"/");
		strcat(cfg.biosJP,biosName);
		if(loadJPBIOS()) {
			return;
		}
		pausVBlank(60*2);
	}
}

void selectGGBios() {
	const char * biosName;
	
	directoryCached = 0;
	biosName = browseForFileType(".gg.zip");
	directoryCached = 0;
	if(biosName) {
		strcpy(cfg.biosGG, currentdir);
		strcat(cfg.biosGG,"/");
		strcat(cfg.biosGG,biosName);
		if(loadGGBIOS()) {
			return;
		}
		pausVBlank(60*2);
	}
}

//---------------------------------------------------------------------------------
const char * browseForFileType(const char * fileTypes) {
	static int pos=0;
	int i, oldpos=-1;
	char * strp;
	int pressed = 0;

	if(fatAvailable) {
		if(dTable == 0)
			dTable = malloc(0x8000);
		if(directoryCached == 0) {
			dItems = getDirectory(dTable, currentdir, fileTypes);
			directoryCached = 1;
		} else {
			chdir("/");			// stupid workaround.
			chdir(currentdir);
		}
		if(dItems) {
			cls(0);
			while(1) {
				waitVBlank();
				pressed = getinput();
				pos = getmenupos(pressed, pos, dItems);
				if(pressed & (KEY_A)) {
					strp = dTable + pos*64;
					if( !(strstr(strp,"~")) ) {
						if(strcmp(strp,"..") == 0) {
							i = strlen(currentdir)-strlen(strrchr(currentdir, '/'));
							currentdir[i] = 0;
							currentdir[i+1] = 0;
							currentdir[0] = '/';
						} else {
							if( (strcmp(currentdir,"/")) )
								strcat(currentdir, "/");
							strncat(currentdir, strp, strlen(strp));
						}
						cls(0);
						dItems=getDirectory(dTable, currentdir, fileTypes);
						pos = 0;
						oldpos = -1;
					} else {
						return ((const char*)strp+1);
						break;
					}
				}
				if(oldpos != pos) {
					oldpos = pos;
					drawfilelist(dTable, pos, dItems);
				}
				if(pressed & (KEY_B))
					break;
			}
		}
//		flushDirectoryCache();
	}
	return NULL;
}

void flushDirectoryCache() {
		free(dTable);
		dTable = 0;
		dItems = 0;
		directoryCached = 0;
}

int loadROM(void* dest, const char* fName, const int maxSize) {
	int size = 0;
	FILE* myfile;
	char fileExt[8];

	getFileExtension(fileExt, fName);
	if(strstr(fileExt, ".zip")) {
		if( (loadFiletypeInZip(dest, fName, FILEEXTENSIONS, maxSize)) == 0) {
			size = cenhead.ucsize;
			strncpy(currentfilename,zipFilename,63);
		} else {
			drawtext(zip_err, 23, 0);
		}
	} else if( (myfile = fopen(fName, "r")) ) {
		fseek(myfile, 0, SEEK_END);
		size = ftell(myfile);
		if( size > maxSize ) {
			drawtext("File to large!", 23, 0);
			return 0;
		}
		fseek(myfile, 0, SEEK_SET);
		fread(dest, 1, size, myfile);
		fclose(myfile);
		strncpy(currentfilename,fName,63);
	} else {
		drawtext("Couldn't open file", 23, 0);
	}
	return size;
}

int loadBIOS(void* dest, const char* fPath, const int maxSize) {
	char tempstring[256];
	char * sptr;

	strcpy(tempstring, fPath);
	if( (sptr = strrchr(tempstring, '/')) ) {
		sptr[0] = 0;
		sptr += 1;
		chdir("/");
		chdir(tempstring);
		return loadROM(dest, sptr, maxSize);
	}
	return 0;
}

int loadUSBIOS(void) {
	if(loadBIOS(BIOS_US_Space, cfg.biosUS, 0x40000)) {
		g_BIOSBASE_US = BIOS_US_Space;
		return 1;
	}
	g_BIOSBASE_US = 0;
	return 0;
}

int loadJPBIOS(void) {
	if(loadBIOS(BIOS_JP_Space, cfg.biosJP, 0x2000)) {
		g_BIOSBASE_JP = BIOS_JP_Space;
		return 1;
	}
	g_BIOSBASE_JP = 0;
	return 0;
}

int loadGGBIOS(void) {
	if(loadBIOS(BIOS_GG_Space, cfg.biosGG, 0x400)) {
		g_BIOSBASE_GG = BIOS_GG_Space;
		return 1;
	}
	g_BIOSBASE_GG = 0;
	return 0;
}

void loadState() {
	u32* statePtr;
	FILE* myfile;
	char stateName[64];

	if(findFolder(FOLDERNAME)) {
		pausVBlank(60*3);
		return;
	}
	strncpy(stateName,currentfilename,59);
	strcat(stateName,".sta");
	if( (statePtr = malloc(STATESIZE)) ) {
		if( (myfile = fopen(stateName, "r")) ) {
			fread(statePtr, 1, STATESIZE, myfile);
			fclose(myfile);
			unpackState(statePtr);
		}
		free(statePtr);
	} else {
		drawtext("Couldn't alloc mem for state.", 18, 0);
		pausVBlank(60*3);
		return;
	}
	drawtext("Loaded state.", 18, 0);
	pausVBlank(60*1);
}
void saveState() {
	u32* statePtr;
	FILE* myfile;
	char stateName[64];

	if(findFolder(FOLDERNAME)) {
		pausVBlank(60*3);
		return;
	}
	strncpy(stateName,currentfilename,59);
	strcat(stateName,".sta");
	if( (statePtr = malloc(STATESIZE)) ) {
		packState(statePtr);
		if( (myfile = fopen(stateName, "w")) ) {
			fwrite(statePtr, 1, STATESIZE, myfile);
			fclose(myfile);
		}
		free(statePtr);
	} else {
		drawtext("Couldn't alloc mem for state.", 18, 0);
		pausVBlank(60*3);
		return;
	}
	drawtext("Saved state.", 18, 0);
	pausVBlank(60*1);
}

int findFolder(const char * folderName) {
	int retValue=0;
	if(!chdir(folderName))
		return retValue;
	chdir("/");
	strcpy(tempdir, "data/");
	strncat(tempdir, folderName, 255);
	if(chdir(tempdir)) {
		if(chdir(folderName)) {
			drawtext("Couldn't find folder:", 18, 0);
			drawtext(folderName, 19, 0);
			retValue = 1;
		}
	}
	return retValue;
}

int getDirectory(char * dirTable, const char * dirName, const char * fileTypes) {
	int i=0;

	char * strP;
	DIR_ITER * pDir;
	struct stat statbuf;
	char name_buffer[MAXPATHLEN];
	char fileExtension[8];

	memset(dirTable,0,0x8000);
	chdir("/");			// stupid workaround.
	chdir(dirName);

	pDir=diropen(dirName);

	drawtext(dirName,0,0);
	if (pDir){
		while (dirnext(pDir, name_buffer, &statbuf) == 0) {
			strP = dirTable + i*64;
			if(strcmp(".", name_buffer) == 0)
				continue;
			if(statbuf.st_mode & S_IFDIR) {
				strncpy(strP, name_buffer,63);
			} else {
				getFileExtension(fileExtension, name_buffer);
				if(fileExtension[0] == 0)
					continue;
				if(strstr(fileTypes, fileExtension) == 0)
					continue;
				strcpy(strP, "~");
				strncat(strP, name_buffer, 63-1);
			}
			i++;
			if(i>511) break;
			drawItem(spinner[i&3], 15, 11, 0);
		}
		dirclose(pDir);
		sortDirectory(dirTable, i-1);
	}
	return i;
}

void sortDirectory(char * dirTable, int count) {
	int i, j=0, dirT;
	char temp[64];

	for(j=0;j<count;j++) {
		dirT = 0;
		for(i=0;i<count;i++) {
			if( strcasecmp(dirTable + i*64, dirTable + (i+1)*64) > 0) {
				strcpy(temp, dirTable + i*64);
				strcpy(dirTable + i*64, dirTable + (i+1)*64);
				strcpy(dirTable + (i+1)*64, temp);
				dirT = 1;
			}
		}
		if(!dirT) break;
	}
}

void getFileExtension(char * dest, const char * fileName) {
	int i;
	char * strExt;

	dest[0]=0;
	if( (strExt = strrchr(fileName, '.')) == 0)
		return;
	strncpy(dest, strExt, 7);
	//Make file extension lower case
	for(i=1;i<6;i++) {
		if(dest[i]) dest[i] |= 0x20;
	}
	
}