#include <nds.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gui.h"
#include "filehandling.h"
#include "romloading.h"
#include "unzip/unzipnds.h"
#include "emubase.h"
#include "ARMZ80/version.h"
#include "SN76496/version.h"

#define ARRSIZE(xxxx) (sizeof((xxxx))/sizeof((xxxx)[0]))
#define EMUVERSION "V0.1"

//asm variables
extern u8 g_emuflags;			// cart.s
extern u8 g_scaling_set;		// cart.s
extern u8 g_machine_set;		// cart.s
extern u8 g_config_set;			// cart.s
extern u8 g_region;				// cart.s
extern u8 g_flicker;			// gfx.s
extern u8 g_twitch;				// gfx.s
extern u8 bcolor;				// gfx.s border color
extern u8 SPRS;					// gfx.s
extern u8 g_gfxmask;			// gfx.s
extern u8 g_gammavalue;			// gfx.s
extern u8 sprcollision;			// cpu.s
extern u32 joycfg;				// io.s
extern u32 EMUinput;

//asm functions
extern void CPU_reset(void);		// cpu.s
extern void ntsc_pal_reset(void);	// cpu.s
extern void loadcart(int,int);		// cart.s
extern void ejectCart(void);		// cart.s
extern void paletteInit(void);		// gfx.s
extern void paletteTxAll(void);		// gfx.s
extern void makeborder(void);		// gfx.s
extern void scanlinebp_reset(void);	// gfx.s
extern void SetupScaling(void);		// gfx.s
extern void SetVDPMode(void);		// gfx.s
extern void Sound_set_frequency(void);	// sound.s
extern void Sound_reset(void);		// sound.s

extern int pauseEmulation;		// main.c
extern int powerButton;			// main.c
extern int sleepTime;			// main.c
extern uint16* map0sub;			// main.c

extern void waitVBlank(void);	// main.c
extern void setEmuSpeed(int speed);	// main.c

const fptr fnlist0[]={uiDummy};
const fptr fnlist1[]={selectSMSGame,loadState,saveState,saveSettings,ejectGame,powerOnOff,resetGame};
const fptr fnlist2[]={ui4,ui5,ui6,ui7};
const fptr fnlist3[]={uiDummy};
const fptr fnlist4[]={autoBSet,autoASet,controllerSet,swapABSet,rstartset,selectset};
const fptr fnlist5[]={scalingSet,flickSet,brightSet,borderSet,spriteSet,bgrlayerSet,sprlayerSet};
const fptr fnlist6[]={countryset,machineset,ui8,collisionset};
const fptr fnlist7[]={speedSet,fpsSet,sleepSet,autostateSet,autopauseGameSet,powerSaveSet,screenSwapSet};
const fptr fnlist8[]={biosset,selectUSBios,selectJPBios,selectGGBios};
const fptr *const fnlistX[]={fnlist0,fnlist1,fnlist2,fnlist3,fnlist4,fnlist5,fnlist6,fnlist7,fnlist8};
const u8 menuXitems[]={ARRSIZE(fnlist0),ARRSIZE(fnlist1),ARRSIZE(fnlist2),ARRSIZE(fnlist3),ARRSIZE(fnlist4),ARRSIZE(fnlist5),ARRSIZE(fnlist6),ARRSIZE(fnlist7),ARRSIZE(fnlist8)};
const fptr drawuiX[]={uiNull,uiFile,uiOptions,uiAbout,uiController,uiDisplay,uiMachine,uiSettings,uiBios};
const u8 menuXback[]={0,0,0,0,2,2,2,2,6};

int menuItemRow=0;
int selected=0;
int selectedMenu=0;
int selectedMain=0;
int lastMainMenu=1;
int mainUIPosition=0;

u8 autoA=0,autoB=0;				//0=off, 1=on, 2=R
int emuSettings=0x0031;


char *const autotxt[]={"OFF","ON","with R"};
char *const speedtxt[]={"Normal","200%","Max","50%"};
char *const sleeptxt[]={"5min","10min","30min","OFF"};
char *const brightxt[]={"I","II","III","IIII","IIIII"};
char *const ctrltxt[]={"1P","2P"};
char *const bordtxt[]={"Black","Border Color","None"};
char *const disptxt[]={"Scaled 1:1","Scaled to fit","Scaled to aspect"};
char *const flicktxt[]={"No Flicker","Flicker"};
char *const cntrtxt[]={"Auto","US (NTSC)","Europe (PAL)","Japan (NTSC)"};
char *const machtxt[]={"Auto","SG-1000","Master System","Master System 2","Game Gear","Mega Drive"};


void setupGUI() {
	keysSetRepeat(25, 4);	// delay, repeat.
	cls(0);
	uiNull();
	selected = 0;		// menu item 0.
}

void uiNull() {
	cls(1);
	drawtext("      Touch screen or",10,0);
	drawtext("     press L+R for menu.",11,0);
	pauseEmulation = 0;
	if(emuSettings & 2) {
		powerOff(POWER_2D_B);
		if(emuSettings & 0x100) {
			powerOff(PM_BACKLIGHT_TOP);
		} else {
			powerOff(PM_BACKLIGHT_BOTTOM);
		}
	}
	if(emuSettings & 0x100) {
		lcdMainOnBottom();
	}
}

void uiFile() {
	cls(1);
	drawTabs();
	menuItemRow = 0;
	drawMenuItem("Load Game");
	drawMenuItem("Load State");
	drawMenuItem("Save State");
	drawMenuItem("Save Settings");
	drawMenuItem("Eject Game");
	drawMenuItem("Power On/Off");
	drawMenuItem("Reset Game");
}

void uiOptions() {
	cls(1);
	drawTabs();
	menuItemRow = 0;
	drawMenuItem("Controller");
	drawMenuItem("Display");
	drawMenuItem("Machine");
	drawMenuItem("Settings");
}

void uiAbout() {
	cls(1);
	drawTabs();
	drawtext(" S8DS - Sega 8bit emulator",4,0);

	drawtext(" B:      Button 1",6,0);
	drawtext(" A:      Button 2",7,0);
	drawtext(" DPad:   Move character",8,0);
	drawtext(" Start:  Pause button",9,0);
	drawtext(" Select: Reset",10,0);

	drawtext(" S8DS         " EMUVERSION,21,0);
	drawtext(" ARMZ80       " ARMZ80VERSION,22,0);
	drawtext(" ARMSNGG76496 " ARMSNGGVERSION,23,0);
}

void uiController() {
	cls(1);
	drawTabs();
	drawtext(" Controller Settings",3,0);
	menuItemRow = 0;
	drawSubItem("B autofire: ",autotxt[autoB]);
	drawSubItem("A autofire: ",autotxt[autoA]);
	drawSubItem("Controller: ",ctrltxt[(joycfg>>29)&1]);
	drawSubItem("Swap A-B:   ",autotxt[(~joycfg>>10)&1]);
	drawSubItem("Use R as Start: ",autotxt[(g_config_set>>6)&1]);
	drawSubItem("Use Select as Reset: ",autotxt[(g_config_set>>5)&1]);
}

void uiDisplay() {
	cls(1);
	drawTabs();
	drawtext(" Display Settings",3,0);
	menuItemRow = 0;
	drawSubItem("Display: ",disptxt[g_scaling_set]);
	drawSubItem("Scaling: ",flicktxt[g_flicker]);
	drawSubItem("Gamma: ",brightxt[g_gammavalue]);
	drawSubItem("GG Border: ",bordtxt[bcolor]);
	drawSubItem("Perfect Sprites: ",autotxt[SPRS]);
	drawSubItem("Disable Background: ",autotxt[g_gfxmask&1]);
	drawSubItem("Disable Sprites: ",autotxt[(g_gfxmask>>4)&1]);
}

void uiMachine() {
	cls(1);
	drawTabs();
	drawtext(" Machine Settings",3,0);
	menuItemRow = 0;
	drawSubItem("Region: ",cntrtxt[g_region]);
	drawSubItem("Machine: ",machtxt[g_machine_set]);
	drawMenuItem(" Bios Settings");
	drawSubItem("Fake Spritecollision: ",autotxt[(sprcollision>>5)&1]);
}

void uiSettings() {
	cls(1);
	drawTabs();
	drawtext(" Settings",3,0);
	menuItemRow = 0;
	drawSubItem("Speed: ",speedtxt[(emuSettings>>6)&3]);
	drawSubItem("FPS-Meter: ",autotxt[(emuSettings>>3)&1]);
	drawSubItem("Autosleep: ",sleeptxt[(emuSettings>>4)&3]);
	drawSubItem("Autoload State: ",autotxt[(emuSettings>>2)&1]);
	drawSubItem("Autopause Game: ",autotxt[emuSettings&1]);
	drawSubItem("Powersave 2nd Screen: ",autotxt[(emuSettings>>1)&1]);
	drawSubItem("Emulator on Bottom: ",autotxt[(emuSettings>>8)&1]);
}

void uiBios() {
	cls(1);
	drawTabs();
	drawtext(" Bios Settings",3,0);
	menuItemRow = 0;
	drawSubItem("Use BIOS: ",autotxt[(g_config_set>>7)&1]);
	drawMenuItem(" Select Export Bios");
	drawMenuItem(" Select Japanese Bios");
	drawMenuItem(" Select GameGear Bios");
}

char menuTopRow[]={   0x82,0x83, ' ',0x81,0x82,0x82,0x82,0x82,0x83, ' ',0x81,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x83, ' ',0x81,0x82,0x82,0x82,0x82,0x82,0x83,0};
char menuMiddleRow[]={ 'X',0x85, ' ',0x84, 'F', 'i', 'l', 'e',0x85, ' ',0x84, 'O', 'p', 't', 'i', 'o', 'n', 's',0x85, ' ',0x84, 'A', 'b', 'o', 'u', 't',0x85,0};
char menuBottomRow[]={ ' ',0x8A, ' ',0x89, ' ', ' ', ' ', ' ',0x8A, ' ',0x89, ' ', ' ', ' ', ' ', ' ', ' ', ' ',0x8A, ' ',0x89, ' ', ' ', ' ', ' ', ' ',0x8A,0};
char menuFrontRow[]={ 0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0};

char tabTopFile[]={0x81,0x82,0x82,0x82,0x82,0x83,0};
char tabMidFile[]={0x84, 'F', 'i', 'l', 'e',0x85,0};
char tabBotFile[]={0x89, ' ', ' ', ' ', ' ',0x8A,0};

char tabTopOptions[]={0x81,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x83,0};
char tabMidOptions[]={0x84, 'O', 'p', 't', 'i', 'o', 'n', 's',0x85,0};
char tabBotOptions[]={0x89, ' ', ' ', ' ', ' ', ' ', ' ', ' ',0x8A,0};

char tabTopAbout[]={0x81,0x82,0x82,0x82,0x82,0x82,0x83,0};
char tabMidAbout[]={0x84, 'A', 'b', 'o', 'u', 't',0x85,0};
char tabBotAbout[]={0x89, ' ', ' ', ' ', ' ', ' ',0x8A,0};

void drawTabs() {
	drawbtext(menuTopRow,0,1);
	drawbtext(menuMiddleRow,1,1);
	drawbtext(menuBottomRow,2,1);
	drawtext(menuFrontRow,2,0);
	if(selectedMain == 1) {
		drawItem(tabTopFile,3,0+32,0);
		drawItem(tabMidFile,3,1+32,0);
		drawItem(tabBotFile,3,2,0);
	} else if (selectedMain == 2) {
		drawItem(tabTopOptions,10,0+32,0);
		drawItem(tabMidOptions,10,1+32,0);
		drawItem(tabBotOptions,10,2,0);
	} else if (selectedMain == 3) {
		drawItem(tabTopAbout,20,0+32,0);
		drawItem(tabMidAbout,20,1+32,0);
		drawItem(tabBotAbout,20,2,0);
	}
}

void uiDummy() {
}

void ui0() {
	changeUI(0);
}
void ui1() {
	changeUI(1);
}
void ui2() {
	changeUI(2);
}
void ui3() {
	changeUI(3);
}
void ui4() {
	changeUI(4);
}
void ui5() {
	changeUI(5);
}
void ui6() {
	changeUI(6);
}
void ui7() {
	changeUI(7);
}
void ui8() {
	changeUI(8);
}
void changeUI(int UInr) {
	selected=0;
	selectedMenu=UInr;
	drawuiX[selectedMenu]();
}


void nullUI() {
	int key;

	key = getinput();
	if( (key&KEY_TOUCH) || ((EMUinput&(KEY_L|KEY_R)) == (KEY_L|KEY_R)) ) {
		selectedMenu = selectedMain = lastMainMenu;
		drawuiX[selectedMenu]();
		powerOn(POWER_2D_B);
		powerOn(PM_BACKLIGHT_TOP);
		powerOn(PM_BACKLIGHT_BOTTOM);
		lcdMainOnTop();
		if(emuSettings&1) {				// Should we pause when menu is open?
			pauseEmulation = 1;
		}
	}
}

void subUI() {
	int key;

	key=getmenuinput(menuXitems[selectedMenu]);
	if(key&(KEY_A)) {
		if(selectedMain == selectedMenu) {		// are we going from a mainMenu to a subMenu?
			mainUIPosition = selected;
		}
		fnlistX[selectedMenu][selected]();
	}
	if(key&(KEY_B)) {
		lastMainMenu = selectedMain;
		selectedMenu = menuXback[selectedMenu];
		if(selectedMain == selectedMenu) {		// are we going back to a mainMenu?
			selected = mainUIPosition;
		}
	}
	if(key&(KEY_L)) {
		selectedMain--;
		if(selectedMain<=0) {
			lastMainMenu = 1;
			selectedMain = 0;
		}
		selectedMenu = selectedMain;
		selected=0;
	}
	if(key&(KEY_R)) {
		selectedMain++;
		if(selectedMain>3) {
			selectedMain = 3;
		} else {
			selectedMenu = selectedMain;
			selected=0;
		}
	}
	if(key&(KEY_A+KEY_B+KEY_L+KEY_R+KEY_UP+KEY_DOWN+KEY_LEFT+KEY_RIGHT+KEY_TOUCH)) {
		drawuiX[selectedMenu]();
	}
}

void zipTest() {
	char filelist[2048];
	listZipFiles("gberet.zip", filelist);
}

int getmenuinput(int menuitems) {
	int keyhit;

	keyhit = getinput();
	selected = getMenuTouch(&keyhit, selected, menuitems);
	selected = getmenupos(keyhit, selected, menuitems);
	return keyhit;
}

//---------------------------------------------------------------------------------
int getinput() {
	int dpad;
	int keyhit=keysDown();	// buttons pressed this loop

	EMUinput = keysHeld();

	dpad = keysDownRepeat()&(KEY_UP+KEY_DOWN+KEY_LEFT+KEY_RIGHT);
	return dpad|(keyhit&(KEY_A+KEY_B+KEY_START+KEY_L+KEY_R+KEY_TOUCH));
}

//---------------------------------------------------------------------------------
int getmenupos(int keyhit, int sel, int menuitems) {
	if(keyhit&KEY_UP)
		sel=(sel+menuitems-1)%menuitems;
	if(keyhit&KEY_DOWN)
		sel=(sel+1)%menuitems;
	if(keyhit&KEY_RIGHT) {
		sel+=10;
		if(sel>menuitems-1) sel=menuitems-1;
	}
	if(keyhit&KEY_LEFT) {
		sel-=10;
		if(sel<0) sel=0;
	}
	return sel;
}

//---------------------------------------------------------------------------------
int getMenuTouch(int* keyhit, int sel, int menuitems) {
	int ypos, xpos;
	touchPosition myTouch;

	if(*keyhit&KEY_TOUCH) {
		touchRead(&myTouch);
		ypos = (myTouch.py>>3) - 5;
		if( (ypos >= 0) && ((ypos>>1) < menuitems) && !(ypos&1) ) {
			if(sel == (ypos>>1) ) {
				*keyhit |= KEY_A;
			} else {
				sel = ypos>>1;
			}
		} else if(myTouch.py < 24) {
			xpos = myTouch.px>>3;
			if( (xpos > 3) && (xpos < 8)) selectedMenu = selectedMain = 1;
			if( (xpos > 10) && (xpos < 18)) selectedMenu = selectedMain = 2;
			if( (xpos > 20) && (xpos < 26)) selectedMenu = selectedMain = 3;
			if( xpos < 2) *keyhit |= KEY_B;
		}
	}
	return sel;
}

int drawfilelist(const char * dtable, int sel, int items) {
	int i,t,firstitem;
	const char * buf;

	firstitem=sel-11;
	if(sel>(items-13)) {
		firstitem=items-23;
	}
	if(firstitem<0) {
		firstitem=0;
	}

	t = firstitem;
	for(i=0;i<23;i++) {
		buf = dtable + t*64;
		if( *buf == '~' )
			drawtext(buf+1,i+1,i==(sel-firstitem)?1:0);
		else
			drawtext(buf,i+1,i==(sel-firstitem)?1:0);
		t++;
	}
	return (sel-firstitem);
}

void drawtext(const char *str,int row,int hilite) {
	u16 *here=map0sub+row*32;
	int i=0;

	*here=hilite?0x012a:0x0120;
	hilite=(hilite<<12)+0x0100;
//	here++;
	while(str[i]>=' ') {
		here[i]=str[i]|hilite;
		i++;
		if(i>31) break;
	}
	for(;i<32;i++)
		here[i]=0x0120;
}

void drawbtext(const char *str,int row,int shadow) {
	u16 *here=map0sub+(row+32)*32;
	int i=0;

	shadow=(shadow<<13)+0x0100;
	while(str[i]>=' ') {
		here[i]=str[i]|shadow;
		i++;
		if(i>31) break;
	}
	for(;i<32;i++)
		here[i]=0x0120;
}

void strmerge(char *dst,const char *src1,const char *src2) {
	if(dst!=src1)
		strcpy(dst,src1);
	strcat(dst,src2);
}

void drawMenuItem(const char *str1) {
	char str[32];
	if(selected==menuItemRow)
		drawItemBackground(str1, menuItemRow*2+5, 0);
	strmerge(str," ",str1);
	drawtext(str,menuItemRow*2+5,selected==menuItemRow);
	menuItemRow++;
}

void drawSubItem(const char *str1,const char *str2) {
	char str[32];
	if(selected==menuItemRow)
		drawItemBackground(str1, menuItemRow*2+5, 1);
	strmerge(str,"  ",str1);
	strmerge(str,str,str2);
	drawtext(str,menuItemRow*2+5,selected==menuItemRow);
	menuItemRow++;
}

void drawItem(const char *str, int col, int row, int pal) {
	u16 *here=map0sub+col+row*32;
	int attrib = (pal<<12)|0x100;
	int i=0;

	while(str[i]>=' ') {
		here[i] = str[i]|attrib;
		i++;
		if(i>31) break;
	}
}

void drawItemBackground(const char *str, int row, int sub) {
	u16 *here=map0sub+(row+31)*32;
	int i=sub;

	here[i+32] = 0x84|0x2100;
	here[i+64] = 0x86|0x2100;
	here[i++]  = 0x81|0x2100;
	while(str[i]>=' ') {
		here[i+32]= 0x80|0x2100;
		here[i+64]= 0x87|0x2100;
		here[i++] = 0x82|0x2100;
		if(i>31) break;
	}
	here[i+32]= 0x80|0x2100;
	here[i+64]=0x87|0x2100;
	here[i++]=0x82|0x2100;
	here[i+32] = 0x85|0x2100;
	here[i+64] = 0x88|0x2100;
	here[i++] = 0x83|0x2100;
	for(;i<32;i++)
		here[i]=0x0120;
}


void cls(int chrmap) {
	int i=0,len=0x400;				//512x256
	u32 *scr=(u32*)map0sub;
	for(;i<len;i++)
		scr[i]=0x01200120;
}

void powerOnOff() {
	powerButton = !powerButton;
	loadcart(0,g_emuflags);
}

void ejectGame() {
	ejectCart();
}

void resetGame() {
	loadcart(0,g_emuflags);
}


void autoASet() {
	autoA++;
	joycfg|=KEY_A+(KEY_A<<16);
	if(autoA==1)
		joycfg&=~KEY_A;
	else if(autoA==2)
		joycfg&=~(KEY_A<<16);
	else
		autoA=0;
}

void autoBSet() {
	autoB++;
	joycfg|=KEY_B+(KEY_B<<16);
	if(autoB==1)
		joycfg&=~KEY_B;
	else if(autoB==2)
		joycfg&=~(KEY_B<<16);
	else
		autoB=0;
}

void controllerSet() {					//see io.s: refreshEMUjoypads
	joycfg^=0x20000000;
}

void swapABSet() {
	joycfg^=0x400;
}

void rstartset() {
	g_config_set^=0x40;
}

void selectset() {
	g_config_set^=0x20;
}


void scalingSet(){
	g_scaling_set++;
	if(g_scaling_set>=3)
		g_scaling_set=0;
	SetupScaling();
	SetVDPMode();
}

void flickSet() {
	g_flicker++;
	if(g_flicker > 1){
		g_flicker=0;
		g_twitch=0;
	}
}

void borderSet() {
	bcolor++;
	if(bcolor>=3)
		bcolor=0;
	makeborder();
}

void spriteSet() {
	SPRS^=1;
}

void brightSet() {
	g_gammavalue++;
	if (g_gammavalue>4) g_gammavalue=0;
	paletteInit();
	paletteTxAll();					//make new palette visible
}

void bgrlayerSet(){
	g_gfxmask^=0x03;
}

void sprlayerSet(){
	g_gfxmask^=0x10;
}


void speedSet() {
	int i;
	i = (emuSettings+0x40) & 0xC0;
	emuSettings = (emuSettings & ~0xC0) | i;
	setEmuSpeed(i>>6);
}

void fpsSet() {
	emuSettings = emuSettings^0x08;
}

void autostateSet() {
	emuSettings = emuSettings^0x04;
}

void sleepSet() {
	int i;
	i = (emuSettings+0x10) & 0x30;
	emuSettings = (emuSettings & ~0x30) | i;
	if(i==0x00)
		sleepTime=60*60*5;			// 5min
	else if(i==0x10)
		sleepTime=60*60*10;			// 10min
	else if(i==0x20)
		sleepTime=60*60*30;			// 30min
	else if(i==0x30)
		sleepTime=0x7F000000;		// 360days...
}

void autopauseGameSet() {
	emuSettings^=0x01;
	pauseEmulation = (emuSettings&1);
}

void powerSaveSet() {
	emuSettings^=0x02;
}

void screenSwapSet() {
	emuSettings^=0x100;
}

void collisionset() {
	sprcollision^=0x20;
}

void countryset() {
	int i;
	g_region = (g_region+1)&3;
	i = (g_region)-1;
	if(i<0) i = 0;
	g_emuflags = (g_emuflags & ~3) | i;
	ntsc_pal_reset();
	Sound_set_frequency();
	scanlinebp_reset();
	SetupScaling();
	SetVDPMode();
}

void machineset() {
	g_machine_set++;
	if(g_machine_set > 5){
		g_machine_set=0;
	}
}

void biosset() {
	g_config_set^=0x80;
}


