#include <nds.h>
#include <fat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <maxmod9.h>
#include "SN76496/sn76496_c.h"
#include "gui.h"
#include "emubase.h"
#include "filehandling.h"

extern mm_word Vbl_Sound_2( mm_word length, mm_addr dest, mm_stream_formats format );
extern void vblirqhandler(void);
extern void convertPalette(void* dest, void* source, int length);	// gfx.s
extern void GFX_init(void);			//gfx.s
extern void antWars(void);			//gfx.s
extern void ejectCart(void);		// cart.s
extern void loadcart(int,int);		// cart.s
extern void run(void);				// cpu.s
extern void Sound_init(void);		// sound.s

extern u8 waitMaskIn;				// cpu.s
extern u8 waitMaskOut;				// cpu.s
extern u8 ROM_Space[];				// cart.s
extern int rawfont[];				// gfx.s
extern u32 EMUinput;				// gfx.s

extern int selectedMenu;			// gui.c
extern int fatAvailable;			// filehandling.c

void waitVBlank(void);
void checkTimeOut(void);
void setup_graphics(void);
void setup_stream(void);


#define sample_rate  32768
#define buffer_size  512
int pauseEmulation = 0;
int powerButton = 0;
int sleepTime = 0x7FFFFFFF;			// 5min
int sleepTimer = 0x7FFFFFFF;			// 5min
int selectedGame = 0;

mm_ds_system sys;
mm_stream mystream;

uint16* map0sub;
uint16* map1sub;

const fptr fnmain[]={nullUI,subUI,subUI,subUI,subUI,subUI,subUI,subUI,subUI};

u8 guiPalette[]={
	0x00,0x00,0xC0, 0x00,0x00,0x00, 0x81,0x81,0x81, 0x93,0x93,0x93, 0xA5,0xA5,0xA5, 0xB7,0xB7,0xB7, 0xC9,0xC9,0xC9, 0xDB,0xDB,0xDB,
	0xED,0xED,0xED, 0xFF,0xFF,0xFF, 0x00,0x00,0xC0, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00,
	0x00,0x00,0x00, 0x00,0x00,0x00, 0x50,0x78,0x78, 0x60,0x90,0x90, 0x78,0xB0,0xB0, 0x88,0xC8,0xC8, 0x90,0xE0,0xE0, 0xA0,0xF0,0xF0,
	0xB8,0xF8,0xF8, 0xEF,0xFF,0xFF, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00,
	0x00,0x00,0x00, 0x00,0x00,0x00, 0x21,0x21,0x21, 0x33,0x33,0x33, 0x45,0x45,0x45, 0x47,0x47,0x47, 0x59,0x59,0x59, 0x6B,0x6B,0x6B,
	0x7D,0x7D,0x7D, 0x8F,0x8F,0x8F, 0x20,0x20,0xE0, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00,
	0x00,0x00,0x00, 0x00,0x00,0x00, 0x81,0x81,0x81, 0x93,0x93,0x93, 0xA5,0xA5,0xA5, 0xB7,0xB7,0xB7, 0xC9,0xC9,0xC9, 0xDB,0xDB,0xDB,
	0xED,0xED,0xED, 0xFF,0xFF,0xFF, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00, 0x00,0x00,0x00
	};

//---------------------------------------------------------------------------------
void myVblank(void) {
//---------------------------------------------------------------------------------
	VBLANK_INTR_WAIT_FLAGS |= IRQ_VBLANK;
	vblirqhandler();
}

//---------------------------------------------------------------------------------
int main(void) {
//---------------------------------------------------------------------------------
	setup_graphics();
	Sound_init();
	GFX_init();

	setup_stream();
	irqSet(IRQ_VBLANK, myVblank);
	setupGUI();
	ejectCart();
	loadcart(0,0);
	if( !(fatAvailable = fatInitDefault()) ) {
		drawtext("fatInitDefault() failure.",23,0);
	} else {
		initFileHandling();
		loadSettings();
		loadUSBIOS();
		loadJPBIOS();
		loadGGBIOS();
	}
	getinput();

	while (1) {
		waitVBlank();
		checkTimeOut();
		fnmain[selectedMenu]();
		if(powerButton) {
			if(!pauseEmulation) {
				run();
			}
		} else {
			antWars();
		}
//		mmStreamUpdate();
	}    
	return 0;
}

//---------------------------------------------------------------------------------
void pausVBlank(int count) {
//---------------------------------------------------------------------------------
	while(--count)
		waitVBlank();
}

//---------------------------------------------------------------------------------
void waitVBlank() {
//---------------------------------------------------------------------------------
	// Workaround for bug in Bios.
	if(	!(VBLANK_INTR_WAIT_FLAGS & IRQ_VBLANK) ) {
		swiIntrWait(0, IRQ_VBLANK);
	}
	VBLANK_INTR_WAIT_FLAGS &= ~IRQ_VBLANK;
}

//---------------------------------------------------------------------------------
void checkTimeOut() {
//---------------------------------------------------------------------------------
	if(EMUinput) {
		sleepTimer = sleepTime;
	} else {
		sleepTimer--;
		if(sleepTimer < 0) {
			sleepTimer = sleepTime;
			// systemSleep doesn't work as expected.
			//systemSleep();	
		}
	}
}

//---------------------------------------------------------------------------------
void setEmuSpeed(int speed) {
//---------------------------------------------------------------------------------
	if(speed == 0) {			//Normal Speed
		waitMaskIn = 0x01;
		waitMaskOut = 0x00;
	} else if(speed == 1) {		// Double speed
		waitMaskIn = 0x00;
		waitMaskOut = 0x00;
	} else if(speed == 2) {		// Max speed (4x)
		waitMaskIn = 0x00;
		waitMaskOut = 0x01;
	} else if(speed == 3) {		// 50% speed
		waitMaskIn = 0x03;
		waitMaskOut = 0x00;
	}
}

//---------------------------------------------------------------------------------
void setup_graphics() {
//---------------------------------------------------------------------------------
	int i;
	int ix = 0;
	int iy = 0;
	uint16* map0 = (uint16*)SCREEN_BASE_BLOCK(1);
	uint16* map1 = (uint16*)SCREEN_BASE_BLOCK(2);
	map0sub = (uint16*)SCREEN_BASE_BLOCK_SUB(0);
	map1sub = (uint16*)SCREEN_BASE_BLOCK_SUB(1);


	//set up the main display
	videoSetMode(MODE_0_2D | 
					DISPLAY_SPR_1D_LAYOUT | 
					DISPLAY_SPR_ACTIVE | 
					DISPLAY_BG0_ACTIVE |
					DISPLAY_BG1_ACTIVE |
					DISPLAY_BG2_ACTIVE |
					DISPLAY_BG3_ACTIVE |
					DISPLAY_WIN0_ON |
					DISPLAY_WIN1_ON );
	
	//set up the sub display
	videoSetModeSub(MODE_0_2D | 
					DISPLAY_SPR_1D_LAYOUT | 
					DISPLAY_SPR_ACTIVE | 
					DISPLAY_BG0_ACTIVE |
					DISPLAY_BG1_ACTIVE );
	
	//vram banks are somewhat complex
	vramSetMainBanks(VRAM_A_MAIN_BG_0x06000000, VRAM_B_MAIN_SPRITE_0x06400000, VRAM_C_SUB_BG, VRAM_D_SUB_SPRITE);	

	//set up two backgrounds to scroll around
	REG_BG0CNT = BG_COLOR_16 | (2 << MAP_BASE_SHIFT);
	REG_BG1CNT = BG_COLOR_16 | (3 << MAP_BASE_SHIFT);
	REG_BG1HOFS = 0;
	REG_BG1VOFS = 0;

	BG_PALETTE[0] = RGB15(0,0,31);
	BG_PALETTE[1] = RGB15(0,31,0);
	BG_PALETTE[2] = RGB15(31,0,0);

	//set up two backgrounds to scroll around
	REG_BG0CNT_SUB = BG_COLOR_16 | (0 << MAP_BASE_SHIFT);
	REG_BG1CNT_SUB = BG_COLOR_16 | (1 << MAP_BASE_SHIFT);
	REG_BG1HOFS_SUB = 0;
	REG_BG1VOFS_SUB = 0;

	//load the maps with alternating tiles (0,1 for bg0 and 0,2 for bg1)
	for(iy = 0; iy < 32; iy++) {
		for(ix = 0; ix <32; ix++) {
			map0[iy * 32 + ix] = (ix ^ iy) & 1;
			map1[iy * 32 + ix] = ((ix ^ iy) & 1)<<1;
		}
	}

	//fill 2 tiles with different colors
	for(i = 0; i < 32 / 2; i++) {
		BG_GFX[i+16] = 0x1111;
		BG_GFX[i+16+16] = 0x2222;
	}

	memcpy(BG_GFX_SUB+0x1200, rawfont, 4096);
	convertPalette(BG_PALETTE_SUB, guiPalette, 64);
	
}

//---------------------------------------------------------------------------------
void setup_stream(void) {
//---------------------------------------------------------------------------------

	//----------------------------------------------------------------
	// initialize maxmod without any soundbank (unusual setup)
	//----------------------------------------------------------------
//	mm_ds_system sys;
	sys.mod_count 			= 0;
	sys.samp_count			= 0;
	sys.mem_bank			= 0;
	sys.fifo_channel		= FIFO_MAXMOD;
	mmInit( &sys );
	
	//----------------------------------------------------------------
	// open stream
	//----------------------------------------------------------------
//	mm_stream mystream;
	mystream.sampling_rate	= sample_rate;				// sampling rate =
	mystream.buffer_length	= buffer_size;				// buffer length =
//	mystream.callback		= mix_sound;				// set callback function
	mystream.callback		= Vbl_Sound_2;				// set callback function
	mystream.format			= MM_STREAM_16BIT_STEREO;	// format = stereo 16-bit
	mystream.timer			= MM_TIMER0;				// use hardware timer 0
	mystream.manual			= false;					// use manual filling
	mmStreamOpen( &mystream );
		
	//----------------------------------------------------------------
	// when using 'automatic' filling, your callback will be triggered
	// every time half of the wave buffer is processed.
	//
	// so: 
	// 25000 (rate)
	// ----- = ~21 Hz for a full pass, and ~42hz for half pass
	// 1200  (length)
	//----------------------------------------------------------------
	// with 'manual' filling, you must call mmStreamUpdate
	// periodically (and often enough to avoid buffer underruns)
	//----------------------------------------------------------------
}

