#include "nds_asm.h"
#include "equates.h"
#include "ARMZ80/z80mac.h"
#include "armsdtmac.h"

	.arm
	.align 4
	.section .text

	.global empty_R
	.global empty_R_SMS1
	.global empty_W
	.global ram_W
	.global ram_W
	.global ram0_W
	.global ram1_W
	.global ram1S_W
	.global ram2_W
	.global sram0_W
	.global sram1_W
	.global rom_R
	.global Korean_W
	.global Korean_MSX_W
	.global Codemaster_W
	.global rom_W
	.global bytecopy_
	.global memset_
	.global memorr_

	.global mem_W8
	.global mem_R8
	.global translatePCToOffset
	.global Z80_memwritetbl
	.global Z80_memreadtbl
	.global rommap

;@----------------------------------------------------------------------------
bytecopy_:					;@ void bytecopy(u8 *dst,u8 *src,int count)
;@----------------------------------------------------------------------------
	subs r2,r2,#1
	ldrplb r3,[r1,r2]
	strplb r3,[r0,r2]
	bhi bytecopy_
	bx lr

;@----------------------------------------------------------------------------
empty_R:					;@ read bad (IO) address, error.
;@----------------------------------------------------------------------------
	mov r11,r11					;@ No$GBA breakpoint
	mov r0,#0xFF				;@ SMS2, GG & MD
	bx lr
;@----------------------------------------------------------------------------
empty_R_SMS1:				;@ read bad (IO) address, error.
;@----------------------------------------------------------------------------
	mov r11,r11					;@ No$GBA breakpoint
	ldrb r0,[r9,#-1]			;@ SMS (z80pc-1)
	bx lr
;@----------------------------------------------------------------------------
empty_W:					;@ write bad address (error)
;@----------------------------------------------------------------------------
	mov r11,r11					;@ No$GBA debugg
	mov r0,#0xBA
	bx lr
;@----------------------------------------------------------------------------
Korean_MSX_W:				;@ write ROM address (bankswitch)
;@----------------------------------------------------------------------------
	cmp addy,#0x0002
	beq BankSwitch0MSX_W
	cmp addy,#0x0003
	beq BankSwitch1MSX_W
	cmp addy,#0x0000
	beq BankSwitch2MSX_W
	cmp addy,#0x0001
	beq BankSwitch3MSX_W
	b rom_W
;@----------------------------------------------------------------------------
Korean_W:					;@ write ROM address (bankswitch)
;@----------------------------------------------------------------------------
	cmp addy,#0xA000
	beq BankSwitch2_W
	b rom_W
;@----------------------------------------------------------------------------
Codemaster_W:				;@ write ROM address (bankswitch)
;@----------------------------------------------------------------------------
	cmp addy,#0x8000
	beq BankSwitch2C_W
	cmp addy,#0x4000
	beq BankSwitch1C_W
	cmp addy,#0x0000
	beq BankSwitch0C_W
;@----------------------------------------------------------------------------
rom_W:						;@ write ROM address (error)
;@----------------------------------------------------------------------------
	mov r11,r11					;@ No$GBA debugg
	mov r0,#0xB0
	bx lr
;@----------------------------------------------------------------------------

	.arm
	.align 4
	.section .itcm
;@----------------------------------------------------------------------------
mem_W8:						;@ mem write ($0000-$FFFF)
;@----------------------------------------------------------------------------
	and r1,addy,#0xE000
	ldr pc,[pc,r1,lsr#11]		;@ in: addy,r0=val(bits 8-31=?)
	nop							;@ out: r0,r1,r2,addy=?
Z80_memwritetbl:
	.long 0,0,0,0,0,0,0,0		;@ $0000-FFFF
;@----------------------------------------------------------------------------
mem_R8:						;@ mem read ($0000-$FFFF)
;@----------------------------------------------------------------------------
	and r1,addy,#0xE000
	ldr pc,[pc,r1,lsr#11]		;@ in: addy,r0=val(bits 8-31=?)
	nop							;@ out: r0,r1,r2,addy=?
Z80_memreadtbl:
	.long 0,0,0,0,0,0,0,0		;@ $0000-FFFF

;@----------------------------------------------------------------------------
ram_W:						;@ ram write ($C000-$DFFF)
;@----------------------------------------------------------------------------
	bic r1,addy,#0xfe000
	ldr r2,=EMU_RAM
	strb r0,[r2,r1]
	bx lr
;@----------------------------------------------------------------------------
ram0_W:						;@ write ram ($C000-$DFFF)
;@----------------------------------------------------------------------------
	ldr r1,=EMU_RAM-0xC000
	strb r0,[r1,addy]
	bx lr
;@----------------------------------------------------------------------------
ram1_W:						;@ write ram ($E000-$FFFF), mirror.
;@----------------------------------------------------------------------------
	ldr r1,=BankMap4
	ldrb r1,[r1]
	tst r1,#0x08				;@ BIOS?
	ldrne r1,=EMU_RAM-0xE000	;@ Skip forward to write with BankSwitch.
	strneb r0,[r1,addy]
	bxne lr
;@----------------------------------------------------------------------------
ram1S_W:					;@ write ram ($E000-$FFFF), mirror. Bankswitch
;@----------------------------------------------------------------------------
	ldr r1,=EMU_RAM-0xE000
	strb r0,[r1,addy]
	mov r1,#0x00040000
	adds r1,r1,addy,lsl#16
	bxmi lr

	beq BankSwitchR_W			;@ $FFFC
	cmp r1,#0x20000
	bhi BankSwitch2_W			;@ $FFFF
	beq BankSwitch1_W			;@ $FFFE
	bmi BankSwitch0_W			;@ $FFFD
;	mov r11,r11
;	bx lr
;@----------------------------------------------------------------------------
ram2_W:						;@ write ram ($C000-$FFFF), mirrored every 2k.
;@----------------------------------------------------------------------------
	bic r2,addy,#0xF800
	ldr r1,=EMU_RAM
	strb r0,[r1,r2]
	bx lr
;@----------------------------------------------------------------------------
sram0_W:					;@ write sram bank0 ($8000-$BFFF)
;@----------------------------------------------------------------------------
	ldr r1,=EMU_SRAM-0x8000
	strb r0,[r1,addy]
	bx lr
;@----------------------------------------------------------------------------
sram1_W:					;@ write sram bank1 ($8000-$BFFF),($C000-$FFFF)
;@----------------------------------------------------------------------------
	ldr r1,=EMU_SRAM-0x4000
	strb r0,[r1,addy]
	bx lr

;@----------------------------------------------------------------------------
rom_R:						;@ rom read
;@----------------------------------------------------------------------------
	and r1,addy,#0xFC00
	adr r2,rommap
	ldr r1,[r2,r1,lsr#8]
	ldrb r0,[r1,addy]
	bx lr
;@----------------------------------------------------------------------------
memset_:						;@ r0=dest r1=data r2=word count
;@	exit with r0 & r1 unchanged, r2=0
;@----------------------------------------------------------------------------
	subs r2,r2,#1
	strpl r1,[r0,r2,lsl#2]
	bhi memset_
	bx lr
;@----------------------------------------------------------------------------
memorr_:						;@ r0=dest r1=data r2=word count
;@	exit with r0 & r1 unchanged, r2=0, r3 trashed
;@----------------------------------------------------------------------------
	subs r2,r2,#1
	ldrpl r3,[r0,r2,lsl#2]
	orrpl r3,r3,r1
	strpl r3,[r0,r2,lsl#2]
	bhi memorr_
	bx lr
;@----------------------------------------------------------------------------
translatePCToOffset:		;@ in = z80pc, out = r0 bank offset
;@----------------------------------------------------------------------------
	and r0,z80pc,#0xFC00
	ldr r0,[pc,r0,lsr#8]		;@ in: z80pc.
	bx lr						;@ out: r0.
;@----------------------------------------------------------------------------
	;@ z80_memmap_tbl
rommap:
	.space 64*4		;@ $0000-FFFF
