#include <nds.h>
#include <fat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "unzipnds.h"
#include "puff.h"

local_file_hdr ziphead;
central_file_hdr cenhead;
endof_central_hdr endhead;
char zip_err[40];
char zipFilename[256];

// Returns offset of file in zipfile. Error if val < 0.
int locateFileInZip(const char* filename, FILE* myfile) {
	int retval=-1;

	if(	(seekCentralZipHeader(&endhead, myfile)) == 0) {
		while(1){
			if(	(fillCentralZipHeader(&cenhead, myfile)) == 0) {
				fread(zipFilename, 1, cenhead.filename_length, myfile);
				zipFilename[cenhead.filename_length] = 0;
				if(strstr(zipFilename, filename)) {
					retval = cenhead.relative_offset_local_header;
					break;
				}
				fseek(myfile, cenhead.extra_field_length + cenhead.file_comment_length, SEEK_CUR);
			} else {
				retval = -1;		// File not found
				break;
			}
		}
	} else {
		retval = -1;		// File not found
	}

	return retval;
}

int locateFiletypeInZip(const char* fileTypes, FILE* myfile) {
	int retval=-1;
	char * strExt=NULL;

	if(	(seekCentralZipHeader(&endhead, myfile)) == 0) {
		while(1){
			if(	(fillCentralZipHeader(&cenhead, myfile)) == 0) {
				fread(zipFilename, 1, cenhead.filename_length, myfile);
				zipFilename[cenhead.filename_length] = 0;
				if( (strExt = strrchr(zipFilename, '.'))) {
					if(strstr(fileTypes, strExt)) {
						retval = cenhead.relative_offset_local_header;
						break;
					}
				}
				fseek(myfile, cenhead.extra_field_length + cenhead.file_comment_length, SEEK_CUR);
			} else {
				retval = -1;		// File not found
				break;
			}
		}
	} else {
		retval = -1;		// File not found
	}

	return retval;
}

int findFileInZip(const char* zipname, const char* filename) {
	int err=0;
	FILE* myfile;

	if( (myfile = fopen(zipname, "r")) ) {
		if( (locateFileInZip(filename, myfile)) < 0)
			err = 1;		// File not found in zip
		fclose(myfile);
	} else {
		err = 1;		// ZipFile not found
	}
	return err;
}

int loadFileInZip(void* dest, const char* zipname, const char* filename, const int maxSize) {
	int err=0, offset=0, csize, ucsize;
	FILE* myfile;
	unsigned char* src;

	if( (myfile = fopen(zipname, "r")) ) {
		offset = locateFileInZip(filename, myfile);
		if( offset >= 0) {
			fseek(myfile, offset, SEEK_SET);
			if(	(fillLocalZipHeader(&ziphead, myfile)) == 0) {
				csize = cenhead.csize;							// localhead doesn't allways contain all info.
				ucsize = cenhead.ucsize;						// centralhead should allways be available.
				if(ucsize > maxSize) {
					strncpy(zip_err,"File to large!", 32);
					err = 1;
					fclose(myfile);
					return err;
				}
				if( ziphead.compression_method == 8 ) {		// 8 == DEFLATE compression.
					if( (src = malloc(csize)) ) {
						fseek(myfile, ziphead.extra_field_length + ziphead.filename_length, SEEK_CUR);
						fread(src, 1, csize, myfile);
						puff(dest, (unsigned long*)&ucsize, src, (unsigned long*)&csize);
						free(src);
					} else {
						strncpy(zip_err,"Out of memory!", 32);
						err = 1;
					}
				} else if( ziphead.compression_method == 0 ) {		// 0 == STORE, no compression.
					fseek(myfile, ziphead.extra_field_length + ziphead.filename_length, SEEK_CUR);
					fread(dest, 1, ziphead.ucsize, myfile);
				} else {									// Only supports DEFLATE & STORE.
					strncpy(zip_err,"Unsupported compression in zip.", 32);
					err = 1;
				}
			} else {
				strncpy(zip_err,"Couldn't find file in zip.", 32);
				err = 1;
			}
		} else {
			strncpy(zip_err,"Couldn't find file in zip.", 32);
			err = 1;
		}
		fclose(myfile);
	} else {
		strncpy(zip_err,"Couldn't open zip file.", 32);
		err = 1;
	}

	return err;
}

int loadFiletypeInZip(void* dest, const char* zipname, const char* fileTypes, const int maxSize) {
	int err=0, offset=0, csize, ucsize;
	FILE* myfile;
	unsigned char* src;

	if( (myfile = fopen(zipname, "r")) ) {
		offset = locateFiletypeInZip(fileTypes, myfile);
		if( offset >= 0) {
			fseek(myfile, offset, SEEK_SET);
			if(	(fillLocalZipHeader(&ziphead, myfile)) == 0) {
				csize = cenhead.csize;							// localhead doesn't allways contain all info.
				ucsize = cenhead.ucsize;						// centralhead should allways be available.
				if(ucsize > maxSize) {
					strncpy(zip_err,"File to large!", 32);
					err = 1;
					fclose(myfile);
					return err;
				}
				if( ziphead.compression_method == 8 ) {		// 8 == DEFLATE compression.
					if( (src = malloc(csize)) ) {
						fseek(myfile, ziphead.extra_field_length + ziphead.filename_length, SEEK_CUR);
						fread(src, 1, csize, myfile);
						puff(dest, (unsigned long*)&ucsize, src, (unsigned long*)&csize);
						free(src);
					} else {
						strncpy(zip_err,"Out of memory!", 32);
						err = 1;
					}
				} else if( ziphead.compression_method == 0 ) {		// 0 == STORE, no compression.
					fseek(myfile, ziphead.extra_field_length + ziphead.filename_length, SEEK_CUR);
					fread(dest, 1, ziphead.ucsize, myfile);
				} else {									// Only supports DEFLATE & STORE.
					strncpy(zip_err,"Unsupported compression in zip.", 32);
					err = 1;
				}
			} else {
				strncpy(zip_err,"Couldn't find file in zip.", 32);
				err = 1;
			}
		} else {
			strncpy(zip_err,"Couldn't find file in zip.", 32);
			err = 1;
		}
		fclose(myfile);
	} else {
		strncpy(zip_err,"Couldn't open zip file.", 32);
		err = 1;
	}

	return err;
}

int listZipFiles(const char* zipname, char* table) {
	int len;
	FILE* myfile;
	char txtbuf[256];

	if( (myfile = fopen(zipname, "r")) ) {
		while(1){
			if(	(fillLocalZipHeader(&ziphead, myfile)) == 0) {
				fread(txtbuf, 1, ziphead.filename_length, myfile);
				txtbuf[ziphead.filename_length] = 0;
				strcat(table, txtbuf);
				len = ziphead.extra_field_length + ziphead.csize;
				fseek(myfile, len, SEEK_CUR);
			} else {
				break;
			}
		}
		fclose(myfile);
	} else {
		return 1;
	}
	return 0;
}

int seekCentralZipHeader(endof_central_hdr* hdr, FILE* myFile) {
	hdr->magic = 0;
	fseek(myFile, - ENDHEADLEN, SEEK_END);
	fread(hdr, 1, ENDHEADLEN, myFile);
	if(	hdr->magic == ENDSIG) {
		fseek(myFile, hdr->cdir_offset, SEEK_SET);
		return 0;
	} else {
		fseek(myFile, - ENDHEADLEN*2, SEEK_END);
		fread(hdr, 1, ENDHEADLEN, myFile);
		if(	hdr->magic == ENDSIG) {
			fseek(myFile, hdr->cdir_offset, SEEK_SET);
			return 0;
		}
	}
	return -1;
}

int fillLocalZipHeader(local_file_hdr* hdr, FILE* myfile) {
	hdr->magic = 0;
	fread(&hdr->magic, 1, sizeof(u32), myfile);
	fread(&hdr->version_needed_to_extract, 1, sizeof(u16), myfile);
	fread(&hdr->general_purpose_bit_flag, 1, sizeof(u16), myfile);
	fread(&hdr->compression_method, 1, sizeof(u16), myfile);
	fread(&hdr->last_mod_dos_datetime, 1, sizeof(u32), myfile);
	fread(&hdr->crc32, 1, sizeof(u32), myfile);
	fread(&hdr->csize, 1, sizeof(u32), myfile);
	fread(&hdr->ucsize, 1, sizeof(u32), myfile);
	fread(&hdr->filename_length, 1, sizeof(u16), myfile);
	fread(&hdr->extra_field_length, 1, sizeof(u16), myfile);
	if(	hdr->magic == LOCSIG) return 0;
	else return -1;
}

int fillCentralZipHeader(central_file_hdr* hdr, FILE* myfile) {
	hdr->magic = 0;
	fread(&hdr->magic, 1, sizeof(u32), myfile);
	fread(&hdr->version_made_by, 1, sizeof(u16), myfile);
	fread(&hdr->version_needed_to_extract, 1, sizeof(u16), myfile);
	fread(&hdr->general_purpose_bit_flag, 1, sizeof(u16), myfile);
	fread(&hdr->compression_method, 1, sizeof(u16), myfile);
	fread(&hdr->last_mod_dos_datetime, 1, sizeof(u32), myfile);
	fread(&hdr->crc32, 1, sizeof(u32), myfile);
	fread(&hdr->csize, 1, sizeof(u32), myfile);
	fread(&hdr->ucsize, 1, sizeof(u32), myfile);
	fread(&hdr->filename_length, 1, sizeof(u16), myfile);
	fread(&hdr->extra_field_length, 1, sizeof(u16), myfile);
	fread(&hdr->file_comment_length, 1, sizeof(u16), myfile);
	fread(&hdr->disk_number_start, 1, sizeof(u16), myfile);
	fread(&hdr->internal_file_attributes, 1, sizeof(u16), myfile);
	fread(&hdr->external_file_attributes, 1, sizeof(u32), myfile);
	fread(&hdr->relative_offset_local_header, 1, sizeof(u32), myfile);
	if(	hdr->magic == CENSIG) return 0;
	else return -1;
}
