
/* PKZIP header definitions */
#define ZIPMAG 0x4b50			/* two-byte zip lead-in ("PK")*/
#define LOCREM 0x0403			/* remaining two bytes in zip signature */
#define LOCSIG 0x04034b50		/* full signature */
#define CENREM 0x0201			/* remaining two bytes in zip signature */
#define CENSIG 0x02014b50		/* full signature */
#define ENDSIG 0x06054b50		/* full signature */
#define ENDHEADLEN 22			/* End header length */

#define  CRPFLG 1				/*  bit for encrypted entry */
#define  EXTFLG 8				/*  bit for extended local header */

typedef struct pkzip_local_file { /* LOCAL */
    u32 magic;
    u16 version_needed_to_extract;
    u16 general_purpose_bit_flag;
    u16 compression_method;
    u32 last_mod_dos_datetime;
    u32 crc32;
    u32 csize;
    u32 ucsize;
    u16 filename_length;
    u16 extra_field_length;
} local_file_hdr;

typedef struct pkzip_central_file { /* CENTRAL */
    u32 magic;
    u16 version_made_by;
    u16 version_needed_to_extract;
    u16 general_purpose_bit_flag;
    u16 compression_method;
    u32 last_mod_dos_datetime;
    u32 crc32;
    u32 csize;
    u32 ucsize;
    u16 filename_length;
    u16 extra_field_length;
    u16 file_comment_length;
    u16 disk_number_start;
    u16 internal_file_attributes;
    u32 external_file_attributes;
    u32 relative_offset_local_header;
} central_file_hdr;

typedef struct pkzip_endof_central_dir { /* CENTRAL END */
    u32 magic;
    u16 disk_number;
    u16 start_disk;
    u16 entry_count;
    u16 total_entry_count;
    u32 cdir_size;
    u32 cdir_offset;
    u16 comment_length;
} endof_central_hdr;

int locateFileInZip(const char* filename, FILE* myfile);
int locateFiletypeInZip(const char* filetypes, FILE* myfile);
int findFileInZip(const char* zname, const char* fname);
int loadFileInZip(void* dst, const char* zName, const char* fName, const int maxSize);
int loadFiletypeInZip(void* dst, const char* zName, const char* fileTypes, const int maxSize);
int listZipFiles(const char* zname, char* tbl);
int seekCentralZipHeader(endof_central_hdr* hdr, FILE* myFile);
int fillLocalZipHeader(local_file_hdr* hdr, FILE* myFile);
int fillCentralZipHeader(central_file_hdr* hdr, FILE* myFile);

extern char zip_err[];
extern char zipFilename[];
extern central_file_hdr cenhead;
